from unittest import TestCase
from unittest.mock import ANY, Mock, call, patch

import pandas as pd
from mlblocks import MLPipeline

from orion import benchmark
from orion.evaluation import CONTEXTUAL_METRICS as METRICS


def test__sort_leaderboard_rank():
    rank = 'f1'
    metrics = METRICS
    score = pd.DataFrame({
        'pipeline': range(5),
        'f1': range(5),
    })

    expected_return = pd.DataFrame({
        'pipeline': range(5)[::-1],
        'rank': range(1, 6),
        'f1': range(5)[::-1],
    })

    returned = benchmark._sort_leaderboard(score, rank, metrics)
    pd.testing.assert_frame_equal(returned, expected_return)


def test__sort_leaderboard_rank_does_not_exist():
    rank = 'does not exist'
    metrics = {'f1': METRICS['f1']}
    score = pd.DataFrame({
        'pipeline': range(5),
        'f1': range(5),
    })

    expected_return = pd.DataFrame({
        'pipeline': range(5)[::-1],
        'rank': range(1, 6),
        'f1': range(5)[::-1],
    })

    returned = benchmark._sort_leaderboard(score, rank, metrics)
    pd.testing.assert_frame_equal(returned, expected_return)


def test__sort_leaderboard_no_rank():
    rank = None
    metrics = METRICS
    score = {k: range(5) for k in metrics.keys()}
    score['pipeline'] = range(5)
    score = pd.DataFrame(score)

    expected_return = score.iloc[::-1].reset_index(drop=True)
    expected_return['rank'] = range(1, 6)

    returned = benchmark._sort_leaderboard(score, rank, metrics)

    assert len(returned.columns) == len(expected_return.columns)
    assert sorted(returned.columns) == sorted(expected_return.columns)
    pd.testing.assert_frame_equal(returned, expected_return[returned.columns])


def test__detrend_signal_trend():
    df = pd.DataFrame({
        'timestamp': range(5),
        'value': range(5)
    })

    expected_return = pd.DataFrame({
        'timestamp': range(5),
        'value': [0.0] * 5,
    })

    returned = benchmark._detrend_signal(df, 'value')
    pd.testing.assert_frame_equal(returned, expected_return)


def test__detrend_signal_no_trend():
    df = pd.DataFrame({
        'timestamp': range(5),
        'value': [0.0] * 5
    })

    expected_return = df.copy()

    returned = benchmark._detrend_signal(df, 'value')
    pd.testing.assert_frame_equal(returned, expected_return)


def test__get_parameter_pipeline():
    hyperparameters = {
        "pipeline1": "pipeline1.json",
        "pipeline2": "pipeline2.json",
    }
    pipeline = "pipeline1"

    expected_return = "pipeline1.json"
    returned = benchmark._get_parameter(hyperparameters, pipeline)
    assert returned == expected_return


def test__get_parameter_dataset():
    hyperparameters = {
        "dataset1": {
            "pipeline1": "pipeline1.json",
            "pipeline2": "pipeline2.json",
        }
    }
    dataset = "dataset1"

    expected_return = {
        "pipeline1": "pipeline1.json",
        "pipeline2": "pipeline2.json",
    }
    returned = benchmark._get_parameter(hyperparameters, dataset)
    assert returned == expected_return


def test__get_parameter_does_not_exist():
    hyperparameters = None
    pipeline = "pipeline1"

    expected_return = None
    returned = benchmark._get_parameter(hyperparameters, pipeline)
    assert returned == expected_return


class TestBenchmark(TestCase):

    @classmethod
    def setup_class(cls):
        cls.pipeline = Mock(autospec=MLPipeline)
        cls.name = 'pipeline-name'
        cls.dataset = 'dataset-name'
        cls.signal = 'signal-name'
        cls.hyper = None
        cls.distributed = False
        cls.rank = 'metric-name'
        cls.metrics = {
            'metric-name': Mock(autospec=METRICS['f1'], return_value=1)
        }

    def set_score(self, metric, elapsed, holdout):
        return {
            'metric-name': metric,
            'elapsed': elapsed,
            'pipeline': self.name,
            'holdout': holdout,
            'dataset': self.dataset,
            'signal': self.signal,
            'status': 'OK'
        }

    @patch('orion.benchmark.load_anomalies')
    @patch('orion.benchmark.analyze')
    @patch('orion.benchmark._load_pipeline')
    @patch('orion.benchmark.load_signal')
    def test__evaluate_signal(
            self, load_signal_mock, load_pipeline_mock, analyze_mock, load_anomalies_mock):
        train = Mock(autospec=pd.DataFrame)
        test = Mock(autospec=pd.DataFrame)
        load_signal_mock.side_effect = [train, test]
        load_pipeline_mock.return_value = self.pipeline

        anomalies = Mock(autospec=pd.DataFrame)
        analyze_mock.return_value = anomalies

        returned = benchmark._evaluate_signal(
            self.pipeline, self.name, self.dataset, self.signal, self.hyper, self.metrics)

        expected_return = self.set_score(1, ANY, ANY)
        assert returned == expected_return

        expected_calls = [
            call('signal-name-train'),
            call('signal-name-test')
        ]
        assert load_signal_mock.call_args_list == expected_calls

        load_pipeline_mock.assert_called_once_with(self.pipeline, self.hyper)
        analyze_mock.assert_called_once_with(self.pipeline, train, test)
        load_anomalies_mock.assert_called_once_with(self.signal)

    @patch('orion.benchmark.load_anomalies')
    @patch('orion.benchmark.analyze')
    @patch('orion.benchmark._load_pipeline')
    @patch('orion.benchmark.load_signal')
    def test__evaluate_signal_exception(
            self, load_signal_mock, load_pipeline_mock, analyze_mock, load_anomalies_mock):
        train = Mock(autospec=pd.DataFrame)
        test = Mock(autospec=pd.DataFrame)
        load_signal_mock.side_effect = [train, test]
        load_pipeline_mock.return_value = self.pipeline

        analyze_mock.side_effect = Exception("failed analyze.")

        returned = benchmark._evaluate_signal(
            self.pipeline, self.name, self.dataset, self.signal, self.hyper, self.metrics)

        expected_return = self.set_score(0, 0, ANY)
        expected_return['status'] = 'ERROR'
        assert returned == expected_return

        expected_calls = [
            call('signal-name-train'),
            call('signal-name-test')
        ]
        assert load_signal_mock.call_args_list == expected_calls

        load_pipeline_mock.assert_called_once_with(self.pipeline, self.hyper)
        analyze_mock.assert_called_once_with(self.pipeline, train, test)

        assert not load_anomalies_mock.called

    @patch('orion.benchmark.load_anomalies')
    @patch('orion.benchmark.analyze')
    @patch('orion.benchmark._load_pipeline')
    @patch('orion.benchmark.load_signal')
    def test__evaluate_signal_holdout(
            self, load_signal_mock, load_pipeline_mock, analyze_mock, load_anomalies_mock):
        train = Mock(autospec=pd.DataFrame)
        test = Mock(autospec=pd.DataFrame)
        load_signal_mock.side_effect = [train, test]
        load_pipeline_mock.return_value = self.pipeline

        holdout = True

        returned = benchmark._evaluate_signal(
            self.pipeline, self.name, self.dataset, self.signal, self.hyper, self.metrics,
            holdout=holdout)

        expected_return = self.set_score(1, ANY, holdout)
        assert returned == expected_return

        expected_calls = [
            call('signal-name-train'),
            call('signal-name-test')
        ]
        assert load_signal_mock.call_args_list == expected_calls

        load_pipeline_mock.assert_called_once_with(self.pipeline, self.hyper)
        analyze_mock.assert_called_once_with(self.pipeline, train, test)
        load_anomalies_mock.assert_called_once_with(self.signal)

    @patch('orion.benchmark.load_anomalies')
    @patch('orion.benchmark.analyze')
    @patch('orion.benchmark._load_pipeline')
    @patch('orion.benchmark.load_signal')
    def test__evaluate_signal_no_holdout(
            self, load_signal_mock, load_pipeline_mock, analyze_mock, load_anomalies_mock):
        train = test = Mock(autospec=pd.DataFrame)
        load_signal_mock.side_effect = [train, test]
        load_pipeline_mock.return_value = self.pipeline

        holdout = False

        returned = benchmark._evaluate_signal(
            self.pipeline, self.name, self.dataset, self.signal, self.hyper, self.metrics,
            holdout=holdout)

        expected_return = self.set_score(1, ANY, holdout)
        assert returned == expected_return

        expected_calls = [
            call('signal-name')
        ]
        assert load_signal_mock.call_args_list == expected_calls

        load_pipeline_mock.assert_called_once_with(self.pipeline, self.hyper)
        analyze_mock.assert_called_once_with(self.pipeline, train, test)
        load_anomalies_mock.assert_called_once_with(self.signal)

    @patch('orion.benchmark.load_anomalies')
    @patch('orion.benchmark.analyze')
    @patch('orion.benchmark._load_pipeline')
    @patch('orion.benchmark.load_signal')
    def test__evaluate_signal_no_detrend(
            self, load_signal_mock, load_pipeline_mock, analyze_mock, load_anomalies_mock):
        train = Mock(autospec=pd.DataFrame)
        test = Mock(autospec=pd.DataFrame)
        load_signal_mock.side_effect = [train, test]
        load_pipeline_mock.return_value = self.pipeline

        detrend = False

        returned = benchmark._evaluate_signal(
            self.pipeline, self.name, self.dataset, self.signal, self.hyper, self.metrics,
            detrend=detrend)

        expected_return = self.set_score(1, ANY, ANY)
        assert returned == expected_return

        expected_calls = [
            call('signal-name-train'),
            call('signal-name-test')
        ]
        assert load_signal_mock.call_args_list == expected_calls

        load_pipeline_mock.assert_called_once_with(self.pipeline, self.hyper)
        analyze_mock.assert_called_once_with(self.pipeline, train, test)
        load_anomalies_mock.assert_called_once_with(self.signal)

    @patch('orion.benchmark.load_anomalies')
    @patch('orion.benchmark.analyze')
    @patch('orion.benchmark._load_pipeline')
    @patch('orion.benchmark.load_signal')
    @patch('orion.benchmark._detrend_signal')
    def test__evaluate_signal_detrend(self, detrend_signal_mock, load_signal_mock,
                                      load_pipeline_mock, analyze_mock, load_anomalies_mock):

        train = Mock(autospec=pd.DataFrame)
        test = Mock(autospec=pd.DataFrame)
        detrend_signal_mock.side_effect = [train, test]
        load_signal_mock.side_effect = [train, test]
        load_pipeline_mock.return_value = self.pipeline

        detrend = True

        returned = benchmark._evaluate_signal(
            self.pipeline, self.name, self.dataset, self.signal, self.hyper, self.metrics,
            detrend=detrend)

        expected_return = self.set_score(1, ANY, ANY)
        assert returned == expected_return

        expected_calls = [
            call('signal-name-train'),
            call('signal-name-test')
        ]
        assert load_signal_mock.call_args_list == expected_calls

        expected_calls = [
            call(train, 'value'),
            call(test, 'value')
        ]
        assert detrend_signal_mock.call_args_list == expected_calls

        load_pipeline_mock.assert_called_once_with(self.pipeline, self.hyper)
        analyze_mock.assert_called_once_with(self.pipeline, train, test)
        load_anomalies_mock.assert_called_once_with(self.signal)

    @patch('orion.benchmark._evaluate_signal')
    def test__evaluate_pipeline(self, evaluate_signal_mock):
        holdout = (True, False)
        detrend = False

        signals = [self.signal]

        score = self.set_score(1, ANY, ANY)
        evaluate_signal_mock.return_value = score

        benchmark._evaluate_pipeline(
            self.pipeline, self.name, self.dataset, signals, self.hyper, self.metrics,
            self.distributed, holdout, detrend)

        expected_calls = [
            call(self.pipeline, self.name, self.dataset, self.signal,
                 self.hyper, self.metrics, True, detrend),

            call(self.pipeline, self.name, self.dataset, self.signal,
                 self.hyper, self.metrics, False, detrend)
        ]
        assert evaluate_signal_mock.call_args_list == expected_calls

    @patch('orion.benchmark._evaluate_signal')
    def test__evaluate_pipeline_holdout_none(self, evaluate_signal_mock):
        holdout = None
        detrend = False

        signals = [self.signal]

        score = self.set_score(1, ANY, ANY)
        evaluate_signal_mock.return_value = score

        returned = benchmark._evaluate_pipeline(
            self.pipeline, self.name, self.dataset, signals, self.hyper, self.metrics,
            self.distributed, holdout, detrend)

        expected_return = [
            self.set_score(1, ANY, True),
            self.set_score(1, ANY, False)
        ]
        assert returned == expected_return

        expected_calls = [
            call(self.pipeline, self.name, self.dataset, self.signal,
                 self.hyper, self.metrics, True, detrend),

            call(self.pipeline, self.name, self.dataset, self.signal,
                 self.hyper, self.metrics, False, detrend)
        ]
        assert evaluate_signal_mock.call_args_list == expected_calls

    @patch('orion.benchmark._evaluate_signal')
    def test__evaluate_pipeline_holdout(self, evaluate_signal_mock):
        holdout = True
        detrend = False

        signals = [self.signal]

        score = self.set_score(1, ANY, holdout)
        evaluate_signal_mock.return_value = score

        expected_return = [score]
        returned = benchmark._evaluate_pipeline(
            self.pipeline, self.name, self.dataset, signals, self.hyper, self.metrics,
            self.distributed, holdout, detrend)

        assert returned == expected_return

        evaluate_signal_mock.assert_called_once_with(
            self.pipeline, self.name, self.dataset, self.signal, self.hyper, self.metrics,
            holdout, detrend)

    @patch('orion.benchmark._evaluate_signal')
    def test__evaluate_pipeline_no_holdout(self, evaluate_signal_mock):
        holdout = False
        detrend = False

        signals = [self.signal]

        score = self.set_score(1, ANY, holdout)
        evaluate_signal_mock.return_value = score

        expected_return = [score]
        returned = benchmark._evaluate_pipeline(
            self.pipeline, self.name, self.dataset, signals, self.hyper, self.metrics,
            self.distributed, holdout, detrend)

        assert returned == expected_return

        evaluate_signal_mock.assert_called_once_with(
            self.pipeline, self.name, self.dataset, self.signal, self.hyper, self.metrics,
            holdout, detrend)

    @patch('orion.benchmark._evaluate_pipeline')
    def test__evaluate_pipelines(self, evaluate_pipeline_mock):
        holdout = False
        detrend = False

        signals = [self.signal]
        pipelines = {self.name: self.pipeline}

        score = self.set_score(1, ANY, holdout)
        evaluate_pipeline_mock.return_value = [score]

        expected_return = [score]
        returned = benchmark._evaluate_pipelines(pipelines, self.dataset, signals, self.hyper,
                                                 self.metrics, self.distributed, holdout, detrend)

        assert returned == expected_return

        evaluate_pipeline_mock.assert_called_once_with(
            self.pipeline, self.name, self.dataset, signals, self.hyper, self.metrics,
            self.distributed, holdout, detrend)

    @patch('orion.benchmark._evaluate_pipeline')
    def test__evaluate_pipelines_hyperparameter(self, evaluate_pipeline_mock):
        holdout = False
        detrend = False

        signals = [self.signal]
        pipelines = {self.name: self.pipeline}

        hyperparameter = Mock(autospec=dict)
        hyperparameters = {self.name: hyperparameter}

        score = self.set_score(1, ANY, holdout)
        evaluate_pipeline_mock.return_value = [score]

        expected_return = [score]
        returned = benchmark._evaluate_pipelines(pipelines, self.dataset, signals, hyperparameters,
                                                 self.metrics, self.distributed, holdout, detrend)

        assert returned == expected_return

        evaluate_pipeline_mock.assert_called_once_with(
            self.pipeline, self.name, self.dataset, signals, hyperparameter, self.metrics,
            self.distributed, holdout, detrend)

    @patch('orion.benchmark._evaluate_pipelines')
    def test__evaluate_datasets(self, evaluate_pipelines_mock):
        holdout = False
        detrend = False

        signals = [self.signal]
        datasets = {self.dataset: signals}
        pipelines = {self.name, self.pipeline}

        score = self.set_score(1, ANY, holdout)
        evaluate_pipelines_mock.return_value = [score]

        order = ['dataset', 'elapsed', 'holdout', 'metric-name', 'pipeline', 'signal', 'status']
        expected_return = pd.DataFrame.from_records([{
            'metric-name': 1,
            'elapsed': ANY,
            'holdout': holdout,
            'pipeline': self.name,
            'dataset': self.dataset,
            'signal': self.signal,
            'status': 'OK'
        }])[order]

        returned = benchmark._evaluate_datasets(
            pipelines, datasets, self.hyper, self.metrics, self.distributed, holdout, detrend)

        pd.testing.assert_frame_equal(returned, expected_return)

        evaluate_pipelines_mock.assert_called_once_with(
            pipelines, self.dataset, signals, self.hyper, self.metrics,
            self.distributed, holdout, detrend)

    @patch('orion.benchmark._evaluate_datasets')
    def test_benchmark(self, evaluate_datasets_mock):
        signals = [self.signal]
        datasets = {self.dataset: signals}
        pipelines = {self.name, self.pipeline}

        score = self.set_score(1, ANY, ANY)
        evaluate_datasets_mock.return_value = pd.DataFrame.from_records([score])

        order = [
            'pipeline',
            'rank',
            'dataset',
            'elapsed',
            'holdout',
            'metric-name',
            'signal',
            'status']
        expected_return = pd.DataFrame.from_records([{
            'rank': 1,
            'metric-name': 1,
            'elapsed': ANY,
            'holdout': ANY,
            'pipeline': self.name,
            'dataset': self.dataset,
            'signal': self.signal,
            'status': 'OK'
        }])[order]

        returned = benchmark.benchmark(
            pipelines, datasets, self.hyper, self.metrics, self.rank, self.distributed)

        pd.testing.assert_frame_equal(returned, expected_return)

        evaluate_datasets_mock.assert_called_once_with(
            pipelines, datasets, self.hyper, self.metrics, self.distributed, False, False)

    @patch('orion.benchmark._evaluate_datasets')
    def test_benchmark_metrics_list(self, evaluate_datasets_mock):
        holdout = False
        detrend = False

        signals = [self.signal]
        datasets = {self.dataset: signals}
        pipelines = {self.name: self.pipeline}

        metric = Mock(autospec=METRICS['f1'], return_value=1)
        metric.__name__ = 'metric-name'
        metrics = [metric]
        metrics_ = {metric.__name__: metric}

        score = self.set_score(1, ANY, holdout)
        score[metric.__name__] = metric
        evaluate_datasets_mock.return_value = pd.DataFrame.from_records([score])

        order = [
            'pipeline',
            'rank',
            'dataset',
            'elapsed',
            'holdout',
            'metric-name',
            'signal',
            'status']
        expected_return = pd.DataFrame.from_records([{
            'rank': 1,
            'metric-name': metric,
            'elapsed': ANY,
            'holdout': holdout,
            'pipeline': self.name,
            'dataset': self.dataset,
            'signal': self.signal,
            'status': 'OK'
        }])[order]

        returned = benchmark.benchmark(pipelines, datasets, self.hyper, metrics, self.rank,
                                       self.distributed, holdout, detrend)

        pd.testing.assert_frame_equal(returned, expected_return)

        evaluate_datasets_mock.assert_called_once_with(
            pipelines, datasets, self.hyper, metrics_, self.distributed, holdout, detrend)

    @patch('orion.benchmark._evaluate_datasets')
    def test_benchmark_metrics_exception(self, evaluate_datasets_mock):
        holdout = False
        detrend = False

        signals = [self.signal]
        datasets = {self.dataset: signals}
        pipelines = {self.name: self.pipeline}

        metric = 'does-not-exist'
        metrics = [metric]

        score = self.set_score(1, ANY, holdout)
        evaluate_datasets_mock.return_value = pd.DataFrame.from_records([score])

        with self.assertRaises(ValueError) as ex:
            benchmark.benchmark(pipelines, datasets, self.hyper, metrics, self.rank,
                                self.distributed, holdout, detrend)

            self.assertTrue(metric in ex.exception)

    @patch('orion.benchmark._evaluate_datasets')
    def test_benchmark_pipelines_list(self, evaluate_datasets_mock):
        holdout = False
        detrend = False

        signals = [self.signal]
        datasets = {self.dataset: signals}
        pipelines = [self.pipeline]
        pipelines_ = {self.pipeline: self.pipeline}

        score = self.set_score(1, ANY, holdout)
        score['pipeline'] = self.pipeline
        evaluate_datasets_mock.return_value = pd.DataFrame.from_records([score])

        order = [
            'pipeline',
            'rank',
            'dataset',
            'elapsed',
            'holdout',
            'metric-name',
            'signal',
            'status']
        expected_return = pd.DataFrame.from_records([{
            'rank': 1,
            'metric-name': 1,
            'elapsed': ANY,
            'holdout': holdout,
            'pipeline': self.pipeline,
            'dataset': self.dataset,
            'signal': self.signal,
            'status': 'OK'
        }])[order]

        returned = benchmark.benchmark(pipelines, datasets, self.hyper, self.metrics, self.rank,
                                       self.distributed, holdout, detrend)

        pd.testing.assert_frame_equal(returned, expected_return)

        evaluate_datasets_mock.assert_called_once_with(
            pipelines_, datasets, self.hyper, self.metrics, self.distributed, holdout, detrend)



import numpy as np
import pandas as pd
import pytest

from orion.evaluation.point import (
    _point_partition, point_accuracy, point_confusion_matrix, point_f1_score, point_precision,
    point_recall)


@pytest.fixture()
def expected():
    return pd.DataFrame({'timestamp': [3, 4, 5]})


@pytest.fixture()
def observed():
    return pd.DataFrame({'timestamp': [4, 6, 7, 8, 12]})


def test__point_partiton(expected, observed):
    expected = list(expected['timestamp'])
    observed = list(observed['timestamp'])
    expected_parts = [1, 1, 1, 0, 0, 0, 0, 0, 0, 0]
    observed_parts = [0, 1, 0, 1, 1, 1, 0, 0, 0, 1]

    expected_return, observed_return, _ = _point_partition(expected, observed)
    np.testing.assert_array_equal(np.array(expected_return),
                                  np.array(expected_parts))

    np.testing.assert_array_equal(np.array(observed_return),
                                  np.array(observed_parts))


def test_point_confusion_matrix(expected, observed):
    expected_return = (3, 4, 2, 1)
    returned = point_confusion_matrix(expected, observed)
    np.testing.assert_array_equal(np.array(returned),
                                  np.array(expected_return))


def test_point_accuracy(expected, observed):
    expected_return = float(4 / 10)
    returned = point_accuracy(expected, observed)
    assert returned == expected_return


def test_point_precision(expected, observed):
    expected_return = float(1 / 5)
    returned = point_precision(expected, observed)
    assert returned == expected_return


def test_point_recall(expected, observed):
    expected_return = float(1 / 3)
    returned = point_recall(expected, observed)
    assert returned == expected_return


def test_point_f1_score(expected, observed):
    expected_return = float(1 / 4)
    returned = point_f1_score(expected, observed)
    assert returned == expected_return


def test_point_f1_score_nan():
    expected = pd.DataFrame({"timestamp": [2, 3]})
    observed = pd.DataFrame({"timestamp": [4, 5]})
    returned = point_f1_score(expected, observed)
    assert np.isnan(returned)



import logging
from datetime import datetime

from azure.cognitiveservices.anomalydetector import AnomalyDetectorClient
from azure.cognitiveservices.anomalydetector.models import Point, Request
from msrest.authentication import CognitiveServicesCredentials

LOGGER = logging.getLogger(__name__)


def _convert_date(x):
    return datetime.fromtimestamp(x).strftime("%Y-%m-%dT%H:%M:%S.%fZ")


def _convert_anomalies_to_contextual(X, interval=1):
    """ Convert list of timestamps to list of tuples.

    Convert a list of anomalies identified by timestamps,
    to a list of tuples marking the start and end interval
    of anomalies; make it contextually defined.

    Args:
        X (list): contains timestamp of anomalies.
        interval (int): allowed gap between anomalies.

    Returns:
        list:
            tuple (start, end, `None`) timestamp.
    """
    if len(X) == 0:
        return []

    X = sorted(X)

    start_ts = 0
    max_ts = len(X) - 1

    anomalies = list()
    break_point = start_ts
    while break_point < max_ts:
        if X[break_point + 1] - X[break_point] <= interval:
            break_point += 1
            continue

        anomalies.append((X[start_ts], X[break_point], None))
        break_point += 1
        start_ts = break_point

    anomalies.append((X[start_ts], X[break_point], None))
    return anomalies


def split_sequence(X, index, target_column, sequence_size, overlap_size):
    """Split sequences of time series data.

    The function creates a list of input sequences by splitting the input sequence
    into partitions with a specified size and pads it with values from previous
    sequence according to the overlap size.

    Args:
        X (ndarray):
            N-dimensional value sequence to iterate over.
        index (ndarray):
            N-dimensional index sequence to iterate over.
        target_column (int):
            Indicating which column of X is the target.
        sequence_size (int):
            Length of the input sequences.
        overlap_size (int):
            Length of the values from previous window.

    Returns:
        tuple:
            * List of sliced value as ndarray.
            * List of sliced index as ndarray.
    """
    X_ = list()
    index_ = list()

    overlap = 0
    start = 0
    max_start = len(X) - 1

    target = X[:, target_column]

    while start < max_start:
        end = start + sequence_size

        X_.append(target[start - overlap:end])
        index_.append(index[start - overlap:end])

        start = end
        overlap = overlap_size

    return X_, index_


def detect_anomalies(X, index, interval, overlap_size, subscription_key, endpoint, granularity,
                     custom_interval=None, period=None, max_anomaly_ratio=None, sensitivity=None):
    """Microsoft's Azure Anomaly Detection tool.

    Args:
        X (list):
            Array containing the input value sequences.
        index (list):
            Array containing the input index sequences.
        interval (int):
            Integer denoting time span frequency of the data.
        overlap_size (int):
            Length of the values from previous sequence that overlaps with current sequnce.
        subscription_key (str):
            Resource key for authenticating your requests.
        endpoint (str):
            Resource endpoint for sending API requests.
        granularity (str or Granularity):
            Can only be one of yearly, monthly, weekly, daily, hourly or minutely.
            Granularity is used for verify whether input series is valid. Possible values
            include: 'yearly', 'monthly', 'weekly', 'daily', 'hourly', 'minutely'.
        custom_interval (int):
            Integer used to set non-standard time interval, for example, if the series
            is 5 minutes, request can be set as {"granularity":"minutely", "custom_interval":5}.
            If not given, `None` is used.
        period (int):
            Periodic value of a time series. If not given, `None` is used, and the API
            will determine the period automatically.
        max_anomaly_ratio (float):
            Advanced model parameter, max anomaly ratio in a time series. If not given,
            `None` is used.
        sensitivity (int):
            Advanced model parameter, between 0-99, the lower the value is, the larger
            the margin value will be which means less anomalies will be accepted. If not given,
            `None` is used.

        Returns:
            list:
                Array containing start-index, end-index, score for each anomalous sequence.
                Note that the API does not have an anomaly score, and so score is set to `None`.
    """
    client = AnomalyDetectorClient(endpoint, CognitiveServicesCredentials(subscription_key))

    overlap = 0
    result = list()

    for x, idx in zip(X, index):
        series = []
        for i in range(len(x)):
            idx_ = _convert_date(idx[i])
            series.append(Point(timestamp=idx_, value=x[i]))

        request = Request(
            series=series, granularity=granularity, custom_interval=custom_interval, period=period,
            max_anomaly_ratio=max_anomaly_ratio, sensitivity=sensitivity)

        try:
            response = client.entire_detect(request)
            if response.is_anomaly:
                anomalous = response.is_anomaly[overlap:]
                index_ = idx[overlap:]
                result.extend(index_[anomalous])

        except Exception as ex:
            LOGGER.exception(ex)

        overlap = overlap_size

    return _convert_anomalies_to_contextual(result, interval)



import numpy as np


class ThresholdDetector:
    """Threshold Detector.

    This is a dummy anomaly detector that classifies as an anomaly
    any value that is further away from the expected value than
    the maximum distance seen during the fit phase.

    Optionally, it also returns the severity of the anomaly, which
    is simply the difference between the actual value and the maximum
    expected value.

    This detector is here only to serve as reference of what
    an anomaly detection primitive looks like, and is not intended
    to be used in real scenarios.
    """

    def __init__(self, value_column, ratio, severity=True):
        if ratio < 0 or ratio > 1:
            raise ValueError('`ratio` must be a value between 0 and 1')

        self._value_column = value_column
        self._ratio = ratio
        self._severity = severity

    def fit(self, X, y):
        truth = X[self._value_column]
        diffs = np.abs(truth - y)
        max_diff = diffs.max()
        self.threshold_ = self._ratio * max_diff

    def detect(self, X, y):
        truth = X[self._value_column]
        diffs = np.abs(truth - y)
        if not self._severity:
            return (diffs > self.threshold_).astype(int)

        else:
            over_threshold = diffs - self.threshold_
            return np.maximum(over_threshold, np.zeros(len(X)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for the `data` module."""

import os
from unittest.mock import patch

import numpy as np
import pandas as pd

import orion
from orion.data import download, load_signal

DATA_PATH = os.path.join(
    os.path.dirname(os.path.abspath(orion.__file__)),
    'data'
)


# ######## #
# download #
# ######## #
@patch('orion.data.pd.read_csv')
@patch('orion.data.os.path.exists')
def test_download_cached(exists_mock, read_csv_mock):
    # setup
    exists_mock.return_value = True

    # run
    returned = download('a_signal_name')

    # assert
    assert returned == read_csv_mock.return_value

    expected_filename = os.path.join(DATA_PATH, 'a_signal_name.csv')
    read_csv_mock.assert_called_once_with(expected_filename)


@patch('orion.data.pd.read_csv')
@patch('orion.data.os.path.exists')
def test_download_new(exists_mock, read_csv_mock):
    # setup
    exists_mock.return_value = False

    # run
    returned = download('a_signal_name')

    # assert
    assert returned == read_csv_mock.return_value

    expected_url = 'https://d3-ai-orion.s3.amazonaws.com/a_signal_name.csv'
    read_csv_mock.assert_called_once_with(expected_url)

    expected_filename = os.path.join(DATA_PATH, 'a_signal_name.csv')
    returned.to_csv.assert_called_once_with(expected_filename, index=False)


# ########### #
# load_signal #
# ########### #
@patch('orion.data.load_csv')
@patch('orion.data.os.path.isfile')
def test_load_signal_filename(isfile_mock, load_csv_mock):
    # setup
    isfile_mock.return_value = True

    # run
    returned = load_signal('a/path/to/a.csv')

    # assert
    assert returned == load_csv_mock.return_value

    load_csv_mock.assert_called_once_with('a/path/to/a.csv', None, None)


@patch('orion.data.download')
@patch('orion.data.load_csv')
@patch('orion.data.os.path.isfile')
def test_load_signal_nasa_signal_name(isfile_mock, load_csv_mock, lns_mock):
    # setup
    isfile_mock.return_value = False

    # run
    returned = load_signal('S-1')

    # assert
    assert returned == lns_mock.return_value

    load_csv_mock.assert_not_called()
    lns_mock.assert_called_once_with('S-1')


@patch('orion.data.load_csv')
@patch('orion.data.os.path.isfile')
def test_load_signal_test_size(isfile_mock, load_csv_mock):
    # setup
    isfile_mock.return_value = True

    data = pd.DataFrame({
        'timestamp': list(range(10)),
        'value': list(range(10, 20))
    })
    load_csv_mock.return_value = data

    # run
    returned = load_signal('a/path/to/a.csv', test_size=0.33)

    # assert
    assert isinstance(returned, tuple)
    assert len(returned) == 2

    train, test = returned

    expected_train = pd.DataFrame({
        'timestamp': list(range(7)),
        'value': list(np.arange(10, 17).astype(float))
    })

    pd.testing.assert_frame_equal(train, expected_train)

    expected_test = pd.DataFrame({
        'timestamp': list(range(7, 10)),
        'value': list(np.arange(17, 20).astype(float))
    })
    expected_test.index = range(7, 10)
    pd.testing.assert_frame_equal(test, expected_test)



# -*- coding: utf-8 -*-

import ast
import csv
import json
import logging
import os
from datetime import datetime
from functools import partial

import dask
import pandas as pd
from scipy import signal as scipy_signal

from orion.analysis import _load_pipeline, analyze
from orion.data import load_anomalies, load_signal
from orion.evaluation import CONTEXTUAL_METRICS as METRICS
from orion.evaluation import contextual_confusion_matrix

LOGGER = logging.getLogger(__name__)

BENCHMARK_PATH = os.path.join(os.path.join(
    os.path.dirname(os.path.abspath(__file__)), '..'),
    'benchmark'
)

with open(os.path.join(BENCHMARK_PATH, 'datasets.csv'), newline='') as f:
    reader = csv.reader(f)
    BENCHMARK_DATA = {row[0]: ast.literal_eval(row[1]) for row in reader}

with open(os.path.join(BENCHMARK_PATH, 'benchmark_parameters.csv'), newline='') as f:
    reader = csv.reader(f)
    BENCHMARK_PARAMS = {row[0]: ast.literal_eval(row[1]) for row in reader}

PIPELINE_DIR = os.path.join(os.path.dirname(__file__), 'pipelines', 'verified')

VERIFIED_PIPELINES = [
    'arima', 'lstm_dynamic_threshold'
]

VERIFIED_PIPELINES_GPU = {
    'arima': 'arima',
    'lstm_dynamic_threshold': 'lstm_dynamic_threshold_gpu'
}


def _load_signal(signal, holdout):
    if holdout:
        train = load_signal(signal + '-train')
        test = load_signal(signal + '-test')
    else:
        train = test = load_signal(signal)

    return train, test


def _detrend_signal(df, value_column):
    df[value_column] = scipy_signal.detrend(df[value_column])
    return df


def _get_parameter(parameters, name):
    if isinstance(parameters, dict) and name in parameters.keys():
        return parameters[name]

    return None


def _sort_leaderboard(df, rank, metrics):
    if rank not in df.columns:
        rank_ = list(metrics.keys())[0]
        LOGGER.exception("Rank %s is not in %s, using %s instead.",
                         rank, df.columns, rank_)
        rank = rank_

    df.sort_values(rank, ascending=False, inplace=True)
    df.reset_index(drop=True, inplace=True)
    df.index.name = 'rank'
    df.reset_index(drop=False, inplace=True)
    df['rank'] += 1

    return df.set_index('pipeline').reset_index()


def _evaluate_signal(pipeline, name, dataset, signal, hyperparameter, metrics,
                     holdout=True, detrend=False):

    train, test = _load_signal(signal, holdout)

    if detrend:
        train = _detrend_signal(train, 'value')
        test = _detrend_signal(test, 'value')

    try:
        LOGGER.info("Scoring pipeline %s on signal %s (Holdout: %s)",
                    name, signal, holdout)

        pipeline = _load_pipeline(pipeline, hyperparameter)
        start = datetime.utcnow()
        anomalies = analyze(pipeline, train, test)
        elapsed = datetime.utcnow() - start

        truth = load_anomalies(signal)

        scores = {
            name: scorer(truth, anomalies, test)
            for name, scorer in metrics.items()
        }
        scores['elapsed'] = elapsed.total_seconds()
        scores['status'] = 'OK'

    except Exception as ex:
        LOGGER.exception("Exception scoring pipeline %s on signal %s (Holdout: %s), error %s.",
                         name, signal, holdout, ex)

        scores = {
            name: 0 for name in metrics.keys()
        }
        scores['elapsed'] = 0
        scores['status'] = 'ERROR'

    scores['pipeline'] = name
    scores['holdout'] = holdout
    scores['dataset'] = dataset
    scores['signal'] = signal

    return scores


def _evaluate_pipeline(pipeline, pipeline_name, dataset, signals, hyperparameter, metrics,
                       distributed, holdout, detrend):
    if holdout is None:
        holdout = (True, False)
    elif not isinstance(holdout, tuple):
        holdout = (holdout, )

    # hyperparameter settings
    if hyperparameter is None:
        file_path = os.path.join(
            PIPELINE_DIR, pipeline_name, pipeline_name + '_' + dataset.lower() + '.json')
        if os.path.exists(file_path):
            hyperparameter = file_path

    if isinstance(hyperparameter, str) and os.path.exists(hyperparameter):
        LOGGER.info("Loading hyperparameter %s", hyperparameter)
        with open(hyperparameter) as f:
            hyperparameter = json.load(f)

    if distributed:
        function = dask.delayed(_evaluate_signal)
    else:
        function = _evaluate_signal

    scores = list()

    for signal in signals:
        for holdout_ in holdout:
            score = function(pipeline, pipeline_name, dataset, signal, hyperparameter,
                             metrics, holdout_, detrend)

            scores.append(score)

    return scores


def _evaluate_pipelines(pipelines, dataset, signals, hyperparameters, metrics, distributed,
                        holdout, detrend):

    scores = list()
    for name, pipeline in pipelines.items():
        hyperparameter = _get_parameter(hyperparameters, name)
        score = _evaluate_pipeline(pipeline, name, dataset, signals, hyperparameter,
                                   metrics, distributed, holdout, detrend)
        scores.extend(score)

    return scores


def _evaluate_datasets(pipelines, datasets, hyperparameters, metrics, distributed, holdout,
                       detrend):
    delayed = []
    for dataset, signals in datasets.items():
        LOGGER.info("Starting dataset {} with {} signals..".format(
            dataset, len(signals)))

        # dataset configuration
        hyperparameters_ = _get_parameter(hyperparameters, dataset)
        parameters = _get_parameter(BENCHMARK_PARAMS, dataset)
        if parameters is not None:
            detrend, holdout = parameters.values()

        result = _evaluate_pipelines(
            pipelines, dataset, signals, hyperparameters_, metrics, distributed, holdout, detrend)

        delayed.extend(result)

    if distributed:
        persisted = dask.persist(*delayed)
        results = dask.compute(*persisted)

    else:
        results = delayed

    df = pd.DataFrame.from_records(results)
    return df


def _summarize_results(df, metrics):
    """ Summarize the result of benchmark.

    The table is summarized for according to the number of wins each pipeline
    had over ARIMA pipeline per dataset, the number of anomalies detected, and
    the average f1 score acheived by that pipeline.
    """
    def return_cm(x):
        if isinstance(x, int):
            return (0, 0, 0)

        elif len(x) > 3:
            return x[1:]

        return x

    df['confusion_matrix'] = df['confusion_matrix'].apply(return_cm)
    df[['fp', 'fn', 'tp']] = pd.DataFrame(df['confusion_matrix'].tolist(), index=df.index)

    # calculate f1 score
    df = df.groupby(['dataset', 'pipeline'])[['fp', 'fn', 'tp']].sum().reset_index()

    precision = df['tp'] / (df['tp'] + df['fp'])
    recall = df['tp'] / (df['tp'] + df['fn'])
    df['f1'] = 2 * (precision * recall) / (precision + recall)

    result = dict()

    # number of wins over ARIMA
    arima_pipeline = 'arima'
    intermediate = df.set_index(['pipeline', 'dataset'])['f1'].unstack().T
    arima = intermediate.pop(arima_pipeline)

    result['# Wins'] = (intermediate.T > arima).sum(axis=1)
    result['# Wins'][arima_pipeline] = None

    # number of anomalies detected
    result['# Anomalies'] = df.groupby('pipeline')[['tp', 'fp']].sum().sum(axis=1).to_dict()

    # average f1 score
    result['Average F1 Score'] = df.groupby('pipeline')['f1'].mean().to_dict()

    result = pd.DataFrame(result)
    result.index.name = 'pipeline'
    result.reset_index(inplace=True)

    rank = 'Average F1 Score'
    result = _sort_leaderboard(result, rank, metrics)
    result = result.drop('rank', axis=1).set_index('pipeline')

    return result


def benchmark(pipelines=None, datasets=None, hyperparameters=None, metrics=METRICS, rank='f1',
              distributed=False, holdout=False, detrend=False, output_path=None):
    """Evaluate pipelines on the given datasets and evaluate the performance.

    The pipelines are used to analyze the given signals and later on the
    detected anomalies are scored against the known anomalies using the
    indicated metrics.

    Finally, the scores obtained with each metric are averaged accross all the signals,
    ranked by the indicated metric and returned on a ``pandas.DataFrame``.

    Args:
        pipelines (dict or list): dictionary with pipeline names as keys and their
            JSON paths as values. If a list is given, it should be of JSON paths,
            and the paths themselves will be used as names. If not give, all verified
            pipelines will be used for evaluation.
        datasets (dict or list): dictionary of dataset name as keys and list of signals as
            values. If a list is given then it will be under a generic name ``dataset``.
            If not given, all benchmark datasets will be used used.
        hyperparameters (dict or list): dictionary with pipeline names as keys
            and their hyperparameter JSON paths or dictionaries as values. If a list is
            given, it should be of corresponding order to pipelines.
        metrics (dict or list): dictionary with metric names as keys and
            scoring functions as values. If a list is given, it should be of scoring
            functions, and they ``__name__`` value will be used as the metric name.
            If not given, all the available metrics will be used.
        rank (str): Sort and rank the pipelines based on the given metric.
            If not given, rank using the first metric.
        distributed (bool): Whether to use dask for distributed computing. If not given,
            use ``False``.
        holdout (bool): Whether to use the prespecified train-test split. If not given,
            use ``False``.
        detrend (bool): Whether to use ``scipy.detrend``. If not given, use ``False``.
        output_path (str): Location to save the intermediatry results. If not given,
            intermediatry results will not be saved.

    Returns:
        pandas.DataFrame: Table containing the average of the scores obtained with
            each scoring function accross all the signals for each pipeline, ranked
            by the indicated metric.
    """
    pipelines = pipelines or VERIFIED_PIPELINES
    datasets = datasets or BENCHMARK_DATA

    if isinstance(pipelines, list):
        pipelines = {pipeline: pipeline for pipeline in pipelines}

    if isinstance(datasets, list):
        datasets = {'dataset': datasets}

    if isinstance(hyperparameters, list):
        hyperparameters = {pipeline: hyperparameter for pipeline, hyperparameter in
                           zip(pipelines.keys(), hyperparameters)}

    if isinstance(metrics, list):
        metrics_ = dict()
        for metric in metrics:
            if callable(metric):
                metrics_[metric.__name__] = metric
            elif metric in METRICS:
                metrics_[metric] = METRICS[metric]
            else:
                raise ValueError('Unknown metric: {}'.format(metric))

        metrics = metrics_

    results = _evaluate_datasets(
        pipelines, datasets, hyperparameters, metrics, distributed, holdout, detrend)

    if output_path:
        LOGGER.info('Saving benchmark report to %s', output_path)
        results.to_csv(output_path)

    return _sort_leaderboard(results, rank, metrics)


def main(cuda=False, distributed=False):
    # output path
    version = "results.csv"
    output_path = os.path.join(BENCHMARK_PATH, 'results', version)

    # metrics
    del METRICS['accuracy']
    METRICS['confusion_matrix'] = contextual_confusion_matrix
    metrics = {k: partial(fun, weighted=False) for k, fun in METRICS.items()}

    # pipelines
    pipelines = VERIFIED_PIPELINES
    if cuda:
        pipelines = VERIFIED_PIPELINES_GPU

    results = benchmark(
        pipelines=pipelines, metrics=metrics, output_path=output_path, distributed=distributed)

    leaderboard = _summarize_results(results, metrics)
    output_path = os.path.join(BENCHMARK_PATH, 'leaderboard.csv')
    leaderboard.to_csv(output_path)


if __name__ == "__main__":
    main()



import logging
import os

import pandas as pd
from mlblocks import MLPipeline

LOGGER = logging.getLogger(__name__)


def _load_pipeline(pipeline, hyperparams=None):
    if isinstance(pipeline, str) and os.path.isfile(pipeline):
        pipeline = MLPipeline.load(pipeline)
    else:
        pipeline = MLPipeline(pipeline)

    if hyperparams is not None:
        pipeline.set_hyperparameters(hyperparams)

    return pipeline


def _run_pipeline(pipeline, train, test):
    LOGGER.debug("Fitting the pipeline")
    pipeline.fit(train)

    LOGGER.debug("Finding events")
    events = pipeline.predict(test)

    LOGGER.debug("%s events found", len(events))
    return events


def _build_events_df(events):
    events = pd.DataFrame(list(events), columns=['start', 'end', 'score'])
    events['start'] = events['start'].astype(int)
    events['end'] = events['end'].astype(int)

    return events


def analyze(pipeline, train, test=None, hyperparams=None):
    if test is None:
        test = train

    if not isinstance(pipeline, MLPipeline):
        pipeline = _load_pipeline(pipeline, hyperparams)

    events = _run_pipeline(pipeline, train, test)

    return _build_events_df(events)



#!/usr/bin/env python
# coding: utf-8

import warnings   # noqa isort:skip
warnings.filterwarnings("ignore")  # noqa isort:skip

import argparse
import getpass
import logging
import os
import sys
from urllib.error import HTTPError

import tabulate

from orion import PIPELINES
from orion.benchmark import benchmark
from orion.data import load_signal
from orion.db.explorer import OrionDBExplorer


def _reset(explorer, args):
    print('WARNING: This will drop the database!')
    name = input('Please enter the database name to confirm: ')
    if name == args.database:
        print('Dropping database {}'.format(name))
        explorer.drop_database()
    else:
        print('Aborting.')


def _add_dataset(explorer, args):
    if args.signal is None:
        args.signal = args.name

    if not args.start or not args.stop:
        path_or_name = args.location or args.name

        try:
            data = load_signal(path_or_name, None, args.timestamp_column, args.value_column)
        except HTTPError:
            print('Unknown signal: {}'.format(path_or_name))
            sys.exit(1)
        else:
            timestamps = data['timestamp']
            if not args.start:
                args.start = timestamps.min()

            if not args.stop:
                args.stop = timestamps.max()

    explorer.add_dataset(
        args.name,
        args.signal,
        args.satellite,
        args.start,
        args.stop,
        args.location,
        args.timestamp_column,
        args.value_column,
        args.user,
    )


def _add_pipeline(explorer, args):
    try:
        explorer.add_pipeline(
            args.name,
            args.path,
            args.user,
        )
    except FileNotFoundError:
        print('File not found: {}'.format(args.path))
    except IsADirectoryError:
        print('File is a directory: {}'.format(args.path))


def _add_comment(explorer, args):
    explorer.add_comment(
        args.event,
        args.text,
        args.user,
    )


def _list(explorer, args):
    name = args.model
    kwargs = {
        name: getattr(args, name)
        for name in args.filters
    }

    method = getattr(explorer, 'get_' + name)
    documents = method(**kwargs)

    if documents.empty:
        print('No {} found'.format(name))
        return

    for delete_ in args.delete:
        del documents[delete_]

    if args.limit:
        documents = documents.head(args.limit)

    if args.output:
        print('Storing results in {}'.format(args.output))
        documents.to_csv(args.output, index=False)
    else:
        print(documents.to_string())


def _run(explorer, args):
    try:
        datarun = explorer.analyze(args.dataset, args.pipeline, args.user)
        print('Datarun id: {}'.format(datarun.id))
    except Exception as ex:
        print("There was an error processing the dataset {}: {}".format(args.dataset, ex))
        sys.exit(1)


def _process(explorer, args):
    pipeline_name = os.path.basename(args.pipeline)[:-5]
    print('Adding pipeline {}'.format(pipeline_name))
    explorer.add_pipeline(pipeline_name, args.pipeline, args.user)

    for path in args.paths:
        try:
            name = os.path.basename(path)
            if not name.endswith('.csv'):
                raise ValueError('Invalid CSV name: {}'.format('.csv'))

            name = name[:-4]

            print('Adding dataset {}'.format(name))
            explorer.add_dataset(
                name,
                name,
                location=path,
                timestamp_column=args.timestamp_column,
                value_column=args.value_column,
                user_id=args.user,
            )

            print('Processing CSV {}'.format(path))
            explorer.analyze(name, pipeline_name, args.user)

        except Exception as ex:
            print("Error processing CSV {}: {}".format(path, ex))


def _evaluate(args):
    if args.all:
        pipelines = PIPELINES
    else:
        pipelines = args.pipeline

    scores = benchmark(pipelines=pipelines, datasets=args.signal, metrics=args.metric,
                       rank=args.rank, holdout=args.holdout)

    if args.output:
        print('Writing results in {}'.format(args.output))
        scores.to_csv(args.output, index=False)

    print(tabulate.tabulate(
        scores,
        showindex=False,
        tablefmt='github',
        headers=scores.columns
    ))


def get_parser():

    # Common Parent - Shared options
    base = argparse.ArgumentParser(add_help=False)
    base.add_argument('-l', '--logfile',
                      help='Name of the logfile. If not given, log to stdout.')
    base.add_argument('-v', '--verbose', action='count', default=0,
                      help='Be verbose. Use -vv for increased verbosity.')

    common = argparse.ArgumentParser(add_help=False, parents=[base])
    common.add_argument('-D', '--database', default='orion',
                        help='Name of the database to connect to. Defaults to "orion"')

    common_user = argparse.ArgumentParser(add_help=False, parents=[common])
    common_user.add_argument('-u', '--user', default=getpass.getuser(), help='User identifier')

    parser = argparse.ArgumentParser(description='Orion Command Line Interface.')
    parser.set_defaults(function=None)

    action = parser.add_subparsers(title='action', dest='action')
    action.required = True

    reset = action.add_parser('reset', help='Reset (drop) the database', parents=[common])
    reset.set_defaults(function=_reset)
    reset.set_defaults(user=None)

    # add
    add = action.add_parser('add', help='Add an object to the database')
    add_model = add.add_subparsers(title='model', dest='model')
    add_model.required = True

    # Add dataset
    add_dataset = add_model.add_parser('dataset', parents=[common_user], help='Add a new dataset')
    add_dataset.set_defaults(function=_add_dataset)

    add_dataset.add_argument('-T', '--timestamp-column', type=int,
                             help='Position of the timestamp column in the CSV,')
    add_dataset.add_argument('-V', '--value-column', type=int,
                             help='Position of the value column in the CSV,')
    add_dataset.add_argument('-s', '--signal',
                             help='Name or ID of the signal. Defaults to the given `name`')
    add_dataset.add_argument('-S', '--satellite',
                             help='Name or ID of the satellite. Defaults to `None`')
    add_dataset.add_argument('--start', type=int, help='Start time, as an integer unix timestamp')
    add_dataset.add_argument('--stop', type=int, help='Stop time, as an integer unix timestamp')
    add_dataset.add_argument('name', help='Name of this dataset')
    add_dataset.add_argument('location', nargs='?', help='path to the CSV file')

    # Add pipeline
    add_pipeline = add_model.add_parser('pipeline', parents=[common_user],
                                        help='Add a new pipeline')
    add_pipeline.set_defaults(function=_add_pipeline)
    add_pipeline.add_argument('name', help='Name of this pipeline')
    add_pipeline.add_argument('path', help='path to the JSON file')

    # Add comment
    add_comment = add_model.add_parser('comment', parents=[common_user],
                                       help='Add a comment to an event')
    add_comment.set_defaults(function=_add_comment)
    add_comment.add_argument('event', help='ID of the event')
    add_comment.add_argument('text', help='Comment text')

    # list
    common_list = argparse.ArgumentParser(add_help=False, parents=[common])
    common_list.add_argument('-o', '--output',
                             help='Dump the output into the given CSV path.')
    common_list.add_argument('-L', '--limit', type=int, help='Limit the number of results')

    list_ = action.add_parser('list', help='List objects from the database')
    list_model = list_.add_subparsers(title='model', dest='model')
    list_model.required = True
    list_.set_defaults(function=_list)
    list_.set_defaults(delete=[])
    list_.set_defaults(filters=[])
    list_.set_defaults(user=None)

    # list datasets
    list_datasets = list_model.add_parser('datasets', parents=[common_list],
                                          help='List datasets')
    list_datasets.set_defaults(model='datasets')

    # list pipelines
    list_pipelines = list_model.add_parser('pipelines', parents=[common_list],
                                           help='List pipelines')
    list_pipelines.set_defaults(model='pipelines')
    list_pipelines.set_defaults(delete=['mlpipeline'])

    # list dataruns
    list_dataruns = list_model.add_parser('dataruns', parents=[common_list],
                                          help='List dataruns')
    list_dataruns.add_argument('-d', '--dataset', help='ID of the dataset')
    list_dataruns.add_argument('-p', '--pipeline', help='ID of the pipeline')
    list_dataruns.set_defaults(model='dataruns')
    list_dataruns.set_defaults(filters=['dataset', 'pipeline'])

    # list events
    list_events = list_model.add_parser('events', parents=[common_list],
                                        help='List found events')
    list_events.add_argument('-d', '--datarun', nargs='?', help='ID of the datarun')
    list_events.set_defaults(model='events')
    list_events.set_defaults(filters=['datarun'])

    # list comments
    list_comments = list_model.add_parser('comments', parents=[common_list],
                                          help='List event comments')
    list_comments.add_argument('-d', '--datarun', nargs='?', help='ID of the datarun')
    list_comments.add_argument('-e', '--event', nargs='?', help='ID of the event')
    list_comments.set_defaults(model='comments')
    list_comments.set_defaults(filters=['datarun', 'event'])

    # Analyze
    run = action.add_parser('run', help='Run a pipeline on a dataset', parents=[common_user])
    run.add_argument('pipeline', help='ID or name of the pipeline')
    run.add_argument('dataset', help='ID of name of the dataset')
    run.set_defaults(function=_run)

    # Process
    process = action.add_parser('process', parents=[common_user],
                                help='Process one or more signal CSV files using a pipeline')
    process.add_argument('-T', '--timestamp-column', type=int,
                         help='Position of the timestamp column in the CSV,')
    process.add_argument('-V', '--value-column', type=int,
                         help='Position of the value column in the CSV,')
    process.add_argument('pipeline', help='Path to the pipeline JSON')
    process.add_argument('paths', nargs='+', help='Paths to the CSV files')
    process.set_defaults(function=_process)

    # Evaluate
    evaluate = action.add_parser('evaluate', parents=[base],
                                 help='Evaluate one or more pipelines on NASA signals')
    evaluate.add_argument('-s', '--signal', action='append',
                          help='Signal to use. Use multiple times for more signals.')
    evaluate.add_argument('-m', '--metric', action='append',
                          help='Metric to use. Use multiple times for more metrics.')
    evaluate.add_argument('-r', '--rank', help='Rank scores based on this metric.')
    evaluate.add_argument('-o', '--output', help='Write the results in the specified CSV file.')
    evaluate.add_argument('--holdout', dest='holdout', action='store_true', default=None,
                          help='Holdout test data during training.')
    evaluate.add_argument('--no-holdout', dest='holdout', action='store_false', default=None,
                          help='Do not holdout test data curing training.')
    group = evaluate.add_mutually_exclusive_group(required=True)
    group.add_argument('-a', '--all', action='store_true', help='Evaluate all known pipelines.')
    group.add_argument('pipeline', default=[], nargs='*',
                       help='Paths to the pipeline JSONs to evaluate')

    return parser


def logging_setup(verbosity=1, logfile=None, logger_name=None):
    logger = logging.getLogger(logger_name)
    log_level = (3 - verbosity) * 10
    fmt = '%(asctime)s - %(process)d - %(levelname)s - %(module)s - %(message)s'
    formatter = logging.Formatter(fmt)
    logger.setLevel(log_level)
    logger.propagate = False

    if logfile:
        file_handler = logging.FileHandler(logfile)
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    else:
        console_handler = logging.StreamHandler()
        console_handler.setLevel(log_level)
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)


def main():
    parser = get_parser()
    args = parser.parse_args()

    logging_setup(args.verbose, args.logfile)

    if args.action == 'evaluate':
        _evaluate(args)
    else:
        explorer = OrionDBExplorer(args.database)

        args.function(explorer, args)



"""Orion DB Explorer model.

This model defines the ``orion.db.explorer.OrionDBExplorer``, which provides
a simple programatic access to creating and reading objects in the Orion Database.
"""
import json
import logging
import os

from gridfs import GridFS
from mlblocks import MLPipeline
from mongoengine import connect
from mongoengine.errors import NotUniqueError
from pymongo.database import Database

from orion.data import load_signal
from orion.db import schema

LOGGER = logging.getLogger(__name__)


class OrionDBExplorer:
    """User interface for the Orion Database.

    This class provides a user-frienly programming interface to
    interact with the Orion database models.

    Args:
        user (str):
            Unique identifier of the user that creates this OrionExporer
            instance. This username or user ID will be used to populate
            the ``created_by`` field of all the objects created in the
            database during this session.
        database (str):
            Name of the MongoDB database to use. Defaults to ``orion``.
        mongodb_config (dict or str):
            A dict or a path to JSON file with additional arguments can be
            passed to provide connection details different than the defaults
            for the MongoDB Database:
                * ``host``: Hostname or IP address of the MongoDB Instance.
                * ``port``: Port to which MongoDB is listening.
                * ``username``: username to authenticate with.
                * ``password``: password to authenticate with.
                * ``authentication_source``: database to authenticate against.

    Examples:
        Simples use case:
        >>> orex = OrionExplorer('my_username')

        Passing a path to a JSON file with connection details.
        >>> orex = OrionExplorer(
        ...      user='my_username',
        ...      database='orion',
        ...      mongodb_config='/path/to/my/mongodb_config.json',
        ... )

        Passing all the possible initialization arguments as a dict:
        >>> mongodb_config = {
        ...      'host': 'localhost',
        ...      'port': 27017,
        ...      'username': 'orion',
        ...      'password': 'secret_password',
        ...      'authentication_source': 'admin',
        ... }
        >>> orex = OrionExplorer(
        ...      user='my_username',
        ...      database='orion',
        ...      mongodb_config=mongodb_config
        ... )
    """

    def __init__(self, user, database='orion', mongodb_config=None):
        """Initiaize this OrionDBExplorer.

        Args:
            user (str):
                Unique identifier of the user that creates this OrionExporer
                instance. This username or user ID will be used to populate
                the ``created_by`` field of all the objects created in the
                database during this session.
            database (str):
                Name of the MongoDB database to use. Defaults to ``orion``.
            mongodb_config (dict or str):
                A dict or a path to JSON file with additional arguments can be
                passed to provide connection details different than the defaults
                for the MongoDB Database:
                    * ``host``: Hostname or IP address of the MongoDB Instance.
                    * ``port``: Port to which MongoDB is listening.
                    * ``username``: username to authenticate with.
                    * ``password``: password to authenticate with.
                    * ``authentication_source``: database to authenticate against.
        """
        if mongodb_config is None:
            mongodb_config = dict()
        elif isinstance(mongodb_config, str):
            with open(mongodb_config) as config_file:
                mongodb_config = json.load(config_file)
        elif isinstance(mongodb_config, dict):
            mongodb_config = mongodb_config.copy()

        self.user = user
        self.database = mongodb_config.pop('database', database)
        self._db = connect(database, **mongodb_config)
        self._fs = GridFS(Database(self._db, self.database))

    def drop_database(self):
        """Drop the database.

        This method is used for development purposes and will
        most likely be removed in the future.
        """
        self._db.drop_database(self.database)

    # ####### #
    # Dataset #
    # ####### #

    def add_dataset(self, name, entity=None):
        """Add a new Dataset object to the database.

        The Dataset needs to be given a name and, optionally, an identitifier,
        name or ID, of the entity which produced the Dataset.

        Args:
            name (str):
                Name of the Dataset
            entity (str):
                Name or Id of the entity which this Dataset is associated to.
                Defaults to ``None``.

        Raises:
            NotUniqueError:
                If a Dataset with the same name and entity values already exists.

        Returns:
            Dataset
        """
        return schema.Dataset.insert(
            name=name,
            entity=entity,
            created_by=self.user
        )

    def get_datasets(self, name=None, entity=None, created_by=None):
        """Query the Datasets collection.

        All the details about the matching Datasets will be returned in
        a ``pandas.DataFrame``.

        All the arguments are optional, so a call without arguments will
        return a table with information about all the Datasets availabe.

        Args:
            name (str):
                Name of the Dataset.
            entity (str):
                Name or Id of the entity which returned Datasets need to be
                associated to.
            created_by (str):
                Unique identifier of the user that created the Datasets.

        Returns:
            pandas.DataFrame
        """
        return schema.Dataset.find(
            as_df_=True,
            name=name,
            entity=entity,
            created_by=created_by
        )

    def get_dataset(self, dataset=None, name=None, entity=None, created_by=None):
        """Get a Dataset object from the database.

        All the arguments are optional but empty queries are not allowed, so at
        least one argument needs to be passed with a value different than ``None``.

        Args:
            dataset (Dataset, ObjectID or str):
                Dataset object (or the corresponding ObjectID, or its string
                representation) that we want to retreive.
            name (str):
                Name of the Dataset.
            entity (str):
                Name or Id of the entity which this Dataset is associated to.
            created_by (str):
                Unique identifier of the user that created the Dataset.

        Raises:
            ValueError:
                If the no arguments are passed with a value different than
                ``None`` or the query resolves to more than one object.

        Returns:
            Dataset
        """
        return schema.Dataset.get(
            dataset=dataset,
            name=name,
            entity=entity,
            created_by=created_by
        )

    # ###### #
    # Signal #
    # ###### #

    def add_signal(self, name, dataset, data_location=None, start_time=None,
                   stop_time=None, timestamp_column=None, value_column=None):
        """Add a new Signal object to the database.

        The signal needs to be given a name and be associated to a Dataset.

        Args:
            name (str):
                Name of the Signal.
            dataset (Dataset or ObjectID or str):
                Dataset object which the created Signal belongs to or the
                corresponding ObjectId.
            data_location (str):
                Path to the CSV containing the Signal data. If the signal is
                one of the signals provided by Orion, this can be omitted and
                the signal will be loaded based on the signal name.
            start_time (int):
                Optional. Minimum timestamp to use for this signal. If not provided
                this defaults to the minimum timestamp found in the signal data.
            stop_time (int):
                Optional. Maximum timestamp to use for this signal. If not provided
                this defaults to the maximum timestamp found in the signal data.
            timestamp_column (int):
                Optional. Index of the timestamp column.
            value_column (int):
                Optional. Index of the value column.

        Raises:
            NotUniqueError:
                If a Signal with the same name already exists for this Dataset.

        Returns:
            Signal
        """
        data_location = data_location or name
        data = load_signal(data_location, None, timestamp_column, value_column)
        timestamps = data['timestamp']
        if not start_time:
            start_time = timestamps.min()

        if not stop_time:
            stop_time = timestamps.max()

        dataset = self.get_dataset(dataset)

        return schema.Signal.insert(
            name=name,
            dataset=dataset,
            start_time=start_time,
            stop_time=stop_time,
            data_location=data_location,
            timestamp_column=timestamp_column,
            value_column=value_column,
            created_by=self.user
        )

    def add_signals(self, dataset, signals_path=None, start_time=None,
                    stop_time=None, timestamp_column=None, value_column=None):
        """Add a multiple Signal objects to the database.

        All the signals will be added to the Dataset using the CSV filename
        as their name.

        Args:
            dataset (Dataset or ObjectID or str):
                Dataset object which the created Signals belongs to or the
                corresponding ObjectId.
            signals_path (str):
                Path to the folder where the signals can be found. All the CSV
                files in this folder will be added.
            start_time (int):
                Optional. Minimum timestamp to use for these signals. If not provided
                this defaults to the minimum timestamp found in the signal data.
            stop_time (int):
                Optional. Maximum timestamp to use for these signals. If not provided
                this defaults to the maximum timestamp found in the signal data.
            timestamp_column (int):
                Optional. Index of the timestamp column.
            value_column (int):
                Optional. Index of the value column.
        """
        for filename in os.listdir(signals_path):
            if filename.endswith('.csv'):
                signal_name = filename[:-4]   # remove from filename .csv
                try:
                    self.add_signal(
                        name=signal_name,
                        dataset=dataset,
                        data_location=os.path.join(signals_path, filename),
                        start_time=start_time,
                        stop_time=stop_time,
                        timestamp_column=timestamp_column,
                        value_column=value_column,
                    )
                except NotUniqueError:
                    pass

    def get_signals(self, name=None, dataset=None, created_by=None):
        """Query the Signals collection.

        All the details about the matching Signals will be returned in
        a ``pandas.DataFrame``.

        All the arguments are optional, so a call without arguments will
        return a table with information about all the Signals availabe.

        Args:
            name (str):
                Name of the Signal.
            dataset (Dataset, ObjectID or str):
                Dataset object (or the corresponding ObjectID, or its string
                representation) to which the Signals have to belong.
            created_by (str):
                Unique identifier of the user that created the Signals.

        Returns:
            pandas.DataFrame
        """
        return schema.Signal.find(
            as_df_=True,
            name=name,
            dataset=dataset,
            created_by=created_by
        )

    def get_signal(self, signal=None, name=None, dataset=None, created_by=None):
        """Get a Signal object from the database.

        All the arguments are optional but empty queries are not allowed, so at
        least one argument needs to be passed with a value different than ``None``.

        Args:
            signal (Signal, ObjectID or str):
                Signal object (or the corresponding ObjectID, or its string
                representation) that we want to retreive.
            name (str):
                Name of the Signal.
            dataset (Dataset, ObjectID or str):
                Dataset object (or the corresponding ObjectID, or its string
                representation) to which the Signal has to belong.
            created_by (str):
                Unique identifier of the user that created the Signals.

        Raises:
            ValueError:
                If the no arguments are passed with a value different than
                ``None`` or the query resolves to more than one object.

        Returns:
            Signal
        """
        return schema.Signal.get(
            signal=signal,
            name=name,
            dataset=dataset,
            created_by=created_by
        )

    # ######## #
    # Template #
    # ######## #

    def add_template(self, name, template=None):
        """Add a new Template object to the database.

        The template can be passed as a name of a registered MLPipeline,
        or as a path to an MLPipeline JSON specification, or as a full
        dictionary specification of an MLPipeline or directly as an
        MLPipeline instance.

        If the ``template`` argument is not passed, the given ``name`` will
        be used to load an MLPipeline.

        During this step, apart from the Template object, a new Pipeline object
        using the default hyperparameters and with the same name as the
        Template will also be created.

        Args:
            name (str):
                Name of the Template.
            template (str, dict or MLPipeline):
                Name of the MLBlocks template to load or path to its JSON
                file or dictionary specification or MLPipeline instance.
                If not given, the ``name`` of the template is used.

        Raises:
            NotUniqueError:
                If a Template with the same name already exists.

        Returns:
            Template
        """
        template = template or name
        if isinstance(template, str) and os.path.isfile(template):
            with open(template, 'r') as f:
                template = json.load(f)

        pipeline_dict = MLPipeline(template).to_dict()

        template = schema.Template.insert(
            name=name,
            json=pipeline_dict,
            created_by=self.user
        )
        schema.Pipeline.insert(
            name=name,
            template=template,
            json=pipeline_dict,
            created_by=self.user
        )

        return template

    def get_templates(self, name=None, created_by=None):
        """Query the Templates collection.

        All the details about the matching Templates will be returned in
        a ``pandas.DataFrame``, except for the JSON specification of the
        template, which will be removed from the table for readability.

        In order to access the JSON specification of each Template, please
        retreive them using the ``get_template`` method.

        All the arguments are optional, so a call without arguments will
        return a table with information about all the Templates availabe.

        Args:
            name (str):
                Name of the Template.
            created_by (str):
                Unique identifier of the user that created the Templates.

        Returns:
            pandas.DataFrame
        """
        return schema.Template.find(
            as_df_=True,
            name=name,
            created_by=created_by,
            exclude_=['json']
        )

    def get_template(self, template=None, name=None, created_by=None):
        """Get a Template object from the database.

        All the arguments are optional but empty queries are not allowed, so at
        least one argument needs to be passed with a value different than ``None``.

        Args:
            template (Template, ObjectID or str):
                Template object (or the corresponding ObjectID, or its string
                representation) that we want to retreive.
            name (str):
                Name of the Template.
            created_by (str):
                Unique identifier of the user that created the Templates.

        Raises:
            ValueError:
                If the no arguments are passed with a value different than
                ``None`` or the query resolves to more than one object.

        Returns:
            Template
        """
        return schema.Template.get(
            template=template,
            name=name,
            created_by=created_by
        )

    # ######## #
    # Pipeline #
    # ######## #

    def add_pipeline(self, name, template, hyperparameters):
        """Add a new Pipeline object to the database.

        The Pipeline will consist on a copy of the given Template using
        the indicated hyperparameters.

        The hyperparameters can be passed as a dictionary containing the
        hyperparameter values following the MLBlocks specification format,
        or a path to a JSON file containing the corresponding values.

        Args:
            name (str):
                Name of the Pipeline.
            template (Template or ObjectID or str):
                Template object (or the corresponding ObjectID, or its string
                representation) that we want to use to create this Pipeline.
            hyperparamers (dict or str):
                dict containing the hyperparameter values following the MLBlocks
                specification format, or a path to a JSON file containing the
                corresponding values.

        Raises:
            NotUniqueError:
                If a Pipeline with the same name for this Template already exists.

        Returns:
            Pipeline
        """
        pipeline = self.get_template(template).load()
        if isinstance(hyperparameters, str):
            with open(hyperparameters, 'r') as f:
                hyperparameters = json.load(f)

        pipeline.set_hyperparameters(hyperparameters)

        return schema.Pipeline.insert(
            name=name,
            template=template,
            json=pipeline.to_dict(),
            created_by=self.user
        )

    def get_pipelines(self, name=None, template=None, created_by=None):
        """Query the Pipelines collection.

        All the details about the matching Pipelines will be returned in
        a ``pandas.DataFrame``, except for the JSON specification of the
        pipeline, which will be removed from the table for readability.

        In order to access the JSON specification of each Pipeline, please
        retreive them using the ``get_pipeline`` method.

        All the arguments are optional, so a call without arguments will
        return a table with information about all the Pipelines availabe.

        Args:
            name (str):
                Name of the Pipeline.
            template (Template or ObjectID or str):
                Template object (or the corresponding ObjectID, or its string
                representation) from which the Pipelines have to be derived.
            created_by (str):
                Unique identifier of the user that created the Pipelines.

        Returns:
            pandas.DataFrame
        """
        return schema.Pipeline.find(
            as_df_=True,
            name=name,
            template=template,
            created_by=created_by,
            exclude_=['json']
        )

    def get_pipeline(self, pipeline=None, name=None, template=None, created_by=None):
        """Get a Pipeline object from the database.

        All the arguments are optional but empty queries are not allowed, so at
        least one argument needs to be passed with a value different than ``None``.

        Args:
            pipeline (Template, ObjectID or str):
                Pipeline object (or the corresponding ObjectID, or its string
                representation) that we want to retreive.
            name (str):
                Name of the Pipeline.
            template (Template or ObjectID or str):
                Template object (or the corresponding ObjectID, or its string
                representation) from which the Pipeline has to be derived.
            created_by (str):
                Unique identifier of the user that created the Pipeline.

        Raises:
            ValueError:
                If the no arguments are passed with a value different than
                ``None`` or the query resolves to more than one object.

        Returns:
            Pipeline
        """
        return schema.Pipeline.get(
            pipeline=pipeline,
            name=name,
            template=template,
            created_by=created_by,
        )

    # ########## #
    # Experiment #
    # ########## #

    def add_experiment(self, name, template, dataset, signals=None, project=None):
        """Add a new Experiment object to the database.

        The Experiment will have to be associated to a Template and a Dataset.

        Optionally, a list of Signal objects or the corresponding ObjectIds have to
        can be passed to associate this Experiment to only a subset of Signals from
        the Dataset. In this case, the Signals passed need to be part of the Dataset,
        otherwise an Exception will be raised.

        If no Signals are passed, all the Signals from the Dataset are used.

        A project name can also be passed as a string to group experiments of
        a single project together.

        Args:
            name (str):
                Name of the Experiment.
            template (Template or ObjectID or str):
                Template object (or the corresponding ObjectID, or its string
                representation) that we want to use in this Experiment.
            dataset (Dataset or ObjectID or str):
                Dataset object (or the corresponding ObjectID, or its string
                representation) which will be used for this Experiment.
            signals (list[Signal, ObjectId or str]):
                list of Signals (or their corresponding ObjectIds) to be used for
                this Experiment.
            project (str):
                Name of the project which this Experiment belongs to.

        Raises:
            NotUniqueError:
                If an Experiment with the same name for this Template already exists.

        Returns:
            Experiment
        """
        dataset = self.get_dataset(dataset)

        if not signals:
            signals = dataset.signals
        else:
            for signal in signals:
                if self.get_signal(signal).dataset != dataset:
                    raise ValueError('All Signals must belong to the Dataset')

        return schema.Experiment.insert(
            name=name,
            project=project,
            template=template,
            dataset=dataset,
            signals=signals,
            created_by=self.user
        )

    def get_experiments(self, name=None, template=None, dataset=None,
                        signals=None, project=None, created_by=None):
        """Query the Experiments collection.

        All the details about the matching Experiments will be returned in
        a ``pandas.DataFrame``.

        All the arguments are optional, so a call without arguments will
        return a table with information about all the Experiments availabe.

        Args:
            name (str):
                Name of the Experiment.
            template (Template or ObjectID or str):
                Template that the Experiments must use.
            dataset (Dataset or ObjectID or str):
                Dataset that the Experiments must use.
            signals (list[Signal, ObjectId or str]):
                Signals that the Experiments must use.
            project (str):
                Name of the project which the Experiments must belong to.
            created_by (str):
                Unique identifier of the user that created the Experiments.

        Returns:
            pandas.DataFrame
        """
        return schema.Experiment.find(
            as_df_=True,
            name=name,
            project=project,
            template=template,
            dataset=dataset,
            signals=signals,
            created_by=created_by,
        )

    def get_experiment(self, experiment=None, name=None, project=None, template=None,
                       dataset=None, signals=None, created_by=None):
        """Get an Experiment object from the database.

        All the arguments are optional but empty queries are not allowed, so at
        least one argument needs to be passed with a value different than ``None``.

        Args:
            experiment (Experiment, ObjectID or str):
                Experiment object (or the corresponding ObjectID, or its string
                representation) that we want to retreive.
            name (str):
                Name of the Experiment.
            template (Template or ObjectID or str):
                Template that the Experiment must use.
            dataset (Dataset or ObjectID or str):
                Dataset that the Experiment must use.
            signals (list[Signal, ObjectId or str]):
                Signals that the Experiment must use.
            project (str):
                Name of the project which the Experiment must belong to.
            created_by (str):
                Unique identifier of the user that created the Experiment.

        Raises:
            ValueError:
                If the no arguments are passed with a value different than
                ``None`` or the query resolves to more than one object.

        Returns:
            Experiment
        """
        return schema.Experiment.get(
            experiment=experiment,
            name=name,
            project=project,
            template=template,
            dataset=dataset,
            signals=signals,
            created_by=created_by,
        )

    # ####### #
    # Datarun #
    # ####### #

    def add_datarun(self, experiment, pipeline):
        """Add a new Datarun object to the database.

        The Datarun needs to be associated to an Experiment and a Pipeline.

        Args:
            experiment (Experiment or ObjectID or str):
                Experiment object (or the corresponding ObjectID, or its string
                representation) to which this Datarun belongs.
            pipeline (Pipeline or ObjectID or str):
                Pipeline object (or the corresponding ObjectID, or its string
                representation) used by this Datarun.

        Returns:
            Datarun
        """
        return schema.Datarun.insert(
            experiment=experiment,
            pipeline=pipeline,
        )

    def get_dataruns(self, experiment=None, pipeline=None, status=None):
        """Query the Dataruns collection.

        All the details about the matching Dataruns will be returned in
        a ``pandas.DataFrame``.

        All the arguments are optional, so a call without arguments will
        return a table with information about all the Dataruns availabe.

        Args:
            experiment (Experiment or ObjectID or str):
                Experiment to which the Dataruns must belong.
            pipeline (Pipeline or ObjectID or str):
                Pipeline which the Dataruns must use.
            status (str):
                Status which the Dataruns must have.

        Returns:
            pandas.DataFrame
        """
        return schema.Datarun.find(
            as_df_=True,
            experiment=experiment,
            pipeline=pipeline,
            status=status,
            exclude_=['software_versions'],
        )

    def get_datarun(self, datarun=None, experiment=None, pipeline=None, status=None):
        """Get a Datarun object from the database.

        All the arguments are optional but empty queries are not allowed, so at
        least one argument needs to be passed with a value different than ``None``.

        Args:
            datarun (Datarun, ObjectID or str):
                Datarun object (or the corresponding ObjectID, or its string
                representation) that we want to retreive.
            experiment (Experiment or ObjectID or str):
                Experiment to which the Datarun must belong.
            pipeline (Pipeline or ObjectID or str):
                Pipeline which the Datarun must use.
            status (str):
                Status which the Datarun must have.

        Raises:
            ValueError:
                If the no arguments are passed with a value different than
                ``None`` or the query resolves to more than one object.

        Returns:
            Datarun
        """
        return schema.Datarun.get(
            experiment=experiment,
            pipeline=pipeline,
            status=status,
        )

    # ######### #
    # Signalrun #
    # ######### #

    def add_signalrun(self, datarun, signal):
        """Add a new Signalrun object to the database.

        The Signalrun needs to be associated to a Datarun and a Signal.

        Args:
            datarun (Datarun or ObjectID or str):
                Datarun object (or the corresponding ObjectID, or its string
                representation) to which this Signalrun belongs.
            signal (Signal or ObjectID or str):
                Signal object (or the corresponding ObjectID, or its string
                representation) used by this Signalrun.

        Returns:
            Datarun
        """
        return schema.Signalrun.insert(
            datarun=datarun,
            signal=signal,
        )

    def get_signalruns(self, datarun=None, signal=None, status=None):
        """Query the Signalruns collection.

        All the details about the matching Signalruns will be returned in
        a ``pandas.DataFrame``.

        All the arguments are optional, so a call without arguments will
        return a table with information about all the Dataruns availabe.

        Args:
            datarun (Datarun or ObjectID or str):
                Datarun to which the Signalruns must belong.
            signal (Signal or ObjectID or str):
                Signal which the Signalruns must use.
            status (str):
                Status which the Signalruns must have.

        Returns:
            pandas.DataFrame
        """
        return schema.Signalrun.find(
            as_df_=True,
            datarun=datarun,
            signal=signal,
            status=status,
        )

    def get_signalrun(self, signalrun=None, datarun=None, signal=None, status=None):
        """Get a Signalrun object from the database.

        All the arguments are optional but empty queries are not allowed, so at
        least one argument needs to be passed with a value different than ``None``.

        Args:
            signalrun (Signalrun, ObjectID or str):
                Signalrun object (or the corresponding ObjectID, or its string
                representation) that we want to retreive.
            datarun (Datarun or ObjectID or str):
                Datarun to which the Signalrun must belong.
            signal (Signal or ObjectID or str):
                Signal which the Signalrun must use.
            status (str):
                Status which the Signalrun must have.

        Raises:
            ValueError:
                If the no arguments are passed with a value different than
                ``None`` or the query resolves to more than one object.

        Returns:
            Signalrun
        """
        return schema.Signalrun.get(
            signalrun=signalrun,
            datarun=datarun,
            signal=signal,
            status=status,
        )

    # ##### #
    # Event #
    # ##### #

    def add_event(self, start_time, stop_time, source, severity=None,
                  signalrun=None, signal=None):
        """Add a new Event object to the database.

        The Event needs to have at least a start_time and a stop_time,
        and be associated to either a Signal or a Signalrun.

        If a Signalrun is given but no Signal is, the created Event will
        be associated to the Signalrun signal value.

        If both a Signalrun and a Signal are given, the Signal must be
        the one used by the Signalrun.

        Args:
            start_time (int):
                Timestamp at which the event starts.
            stop_time (int):
                Timestamp at which the event ends.
            source (str):
                Description of where this Event was created. It must be "orion",
                "shape matching" or "manually created".
            severity (float):
                Severity score value. Optional.
            signalrun (Signalrun or ObjectID or str):
                Signalrun object (or the corresponding ObjectID, or its string
                representation) to which this Event is associated.
            signal (Signal or ObjectID or str):
                Signal object (or the corresponding ObjectID, or its string
                representation) to which this Event is associated.

        Raises:
            ValueError:
                if neither a Signal or a Signalrun are given, or if the Signal
                is not the one used by the Signalrun.

        Returns:
            Event
        """
        if signal is None and signalrun is None:
            raise ValueError('An Event must be associated to either a Signalrun or a Signal')
        if signal is not None and signalrun is not None:
            if self.get_signal(signal) != self.get_signalrun(signalrun).signal:
                raise ValueError('Signal cannot be different than Signalrun.signal')

        return schema.Event.insert(
            signalrun=signalrun,
            signal=signal or signalrun.signal,
            start_time=int(start_time),
            stop_time=int(stop_time),
            severity=severity,
            source=source,
        )

    def get_events(self, signalrun=None, signal=None, source=None):
        """Query the Events collection.

        All the details about the matching Signalruns will be returned in
        a ``pandas.DataFrame``.

        All the arguments are optional, so a call without arguments will
        return a table with information about all the Events availabe.

        Args:
            signalrun (Signalrun or ObjectID or str):
                Signalrun to which the Events must belong.
            signal (Signal or ObjectID or str):
                Signal to which the Events be associated.
            source (str):
                Source from which the Events must come.

        Returns:
            pandas.DataFrame
        """
        return schema.Event.find(
            as_df_=True,
            signalrun=signalrun,
            signal=signal,
            source=source,
        )

    def get_event(self, event=None, signalrun=None, signal=None, source=None):
        """Get an Event object from the database.

        All the arguments are optional but empty queries are not allowed, so at
        least one argument needs to be passed with a value different than ``None``.

        Args:
            event (Event, ObjectID or str):
                Event object (or the corresponding ObjectID, or its string
                representation) that we want to retreive.
            signalrun (Signalrun or ObjectID or str):
                Signalrun to which the Events must belong.
            signal (Signal or ObjectID or str):
                Signal to which the Events be associated.
            source (str):
                Source from which the Events must come.

        Raises:
            ValueError:
                If the no arguments are passed with a value different than
                ``None`` or the query resolves to more than one object.

        Returns:
            Event
        """
        return schema.Event.get(
            event=event,
            signalrun=signalrun,
            signal=signal,
            source=source,
        )

    # ########## #
    # Annotation #
    # ########## #

    def add_annotation(self, event, tag=None, comment=None):
        """Add a new Annotation object to the database.

        The Event needs to be associated with an Event, and can be given a
        ``tag`` and a text comment.

        Args:
            event (Event or ObjectID or str):
                Event object (or the corresponding ObjectID, or its string
                representation) to which this Annotation is associated.
            tag (str):
                Tag of this Annotation.
            comment (str):
                Text comment of this Annotation.

        Returns:
            Annotation
        """
        return schema.Annotation.insert(
            event=event,
            tag=tag,
            comment=comment,
        )

    def get_annotations(self, event=None, tag=None, comment=None, created_by=None):
        """Query the Annotations collection.

        All the details about the matching Annotations will be returned in
        a ``pandas.DataFrame``.

        All the arguments are optional, so a call without arguments will
        return a table with information about all the Annotations availabe.

        Args:
            event (Event or ObjectID or str):
                Event to which the Annotations must belong.
            tag (str):
                Tag which the Annotations must have.
            comment (str):
                Comment which the Annotations must have.
            created_by (str):
                Unique identifier of the user that created the Annotations.

        Returns:
            pandas.DataFrame
        """
        return schema.Annotation.find(
            as_df_=True,
            event=event,
            tag=tag,
            created_by=created_by,
        )

    def get_annotation(self, annotation=None, event=None, tag=None, created_by=None):
        """Get an Event object from the database.

        All the arguments are optional but empty queries are not allowed, so at
        least one argument needs to be passed with a value different than ``None``.

        Args:
            annotation (Annotation, ObjectID or str):
                Annotation object (or the corresponding ObjectID, or its string
                representation) that we want to retreive.
            event (Event or ObjectID or str):
                Event to which the Annotation must belong.
            tag (str):
                Tag which the Annotation must have.
            comment (str):
                Comment which the Annotation must have.
            created_by (str):
                Unique identifier of the user that created the Annotation.

        Raises:
            ValueError:
                If the no arguments are passed with a value different than
                ``None`` or the query resolves to more than one object.

        Returns:
            Annotation
        """
        return schema.Annotation.get(
            annotation=annotation,
            event=event,
            tag=tag,
            created_by=created_by,
        )



# -*- coding: utf-8 -*-

import logging


def logging_setup(verbosity=1, logfile=None, logger_name=None):
    logger = logging.getLogger(logger_name)
    log_level = (3 - verbosity) * 10
    fmt = '%(asctime)s - %(process)d - %(levelname)s - %(module)s - %(message)s'
    formatter = logging.Formatter(fmt)
    logger.setLevel(log_level)
    logger.propagate = False

    if logfile:
        file_handler = logging.FileHandler(logfile)
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    else:
        console_handler = logging.StreamHandler()
        console_handler.setLevel(log_level)
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)



"""Orion Database Schema.

This module contains the classes that define the Orion Database Schema:
    * Dataset
    * Signal
    * Template
    * Pipeline
    * Experiment
    * Datarun
    * Signalrun
    * Event
    * EventInteraction
    * Annotation
"""

import logging
from datetime import datetime

from mlblocks import MLPipeline
from mongoengine import CASCADE, fields
from pip._internal.operations import freeze

from orion.data import load_signal
from orion.db.base import OrionDocument, PipelineField, Status

LOGGER = logging.getLogger(__name__)


class Dataset(OrionDocument):
    """Dataset object.

    A **Dataset** represents a group of Signals that are grouped together under a
    common name, which is usually defined by an external entity.
    """
    name = fields.StringField(required=True)
    entity = fields.StringField()
    created_by = fields.StringField()

    unique_key_fields = ['name', 'entity']

    @property
    def signals(self):
        return Signal.find(dataset=self)


class Signal(OrionDocument):
    """Signal object.

    A Signal belongs to a Dataset and contains all the required details to be
    able to load the observations from a timeseries signal, as well as some
    metadata about it, such as the minimum and maximum timestamps that have to
    be used.
    """
    name = fields.StringField(required=True)
    dataset = fields.ReferenceField(Dataset, reverse_delete_rule=CASCADE)
    start_time = fields.IntField()
    stop_time = fields.IntField()
    data_location = fields.StringField()
    timestamp_column = fields.IntField()
    value_column = fields.IntField()
    created_by = fields.StringField()

    unique_key_fields = ['name', 'dataset']

    def load(self):
        data = load_signal(self.data_location, None, self.timestamp_column, self.value_column)
        if self.start_time:
            data = data[data['timestamp'] >= self.start_time].copy()

        if self.stop_time:
            data = data[data['timestamp'] <= self.stop_time].copy()

        return data


class Template(OrionDocument):
    """Template object.

    A **Template** represents an MLPipeline template from which later on
    Pipelines will be derived to run on Experiments with different hyperparameter
    values.

    The Template includes the complete JSON specification of the MLPipeline it
    represents.
    """
    name = fields.StringField(required=True)
    json = PipelineField(required=True)
    created_by = fields.StringField()

    unique_key_fields = ['name']

    def load(self):
        """Load this Template as an MLPipeline.

        Returns:
            MLPipeline
        """
        return MLPipeline(self.json)

    @property
    def pipelines(self):
        return Pipeline.find(template=self)


class Pipeline(OrionDocument):
    """Pipeline object.

    A **Pipeline** represents an MLPipeline object.
    It is derived from a Template by setting a specific set of
    hyperparameter values.
    """
    name = fields.StringField(required=True)
    template = fields.ReferenceField(Template, reverse_delete_rule=CASCADE)
    json = PipelineField(required=True)
    created_by = fields.StringField()

    unique_key_fields = ['name', 'template']

    def load(self):
        """Load this Pipeline as an MLPipeline.

        Returns:
            MLPipeline
        """
        return MLPipeline(self.json)


class Experiment(OrionDocument):
    """Experiment object.

    An Experiment is associated with a Dataset, a subset of its Signals and a Template.
    It represents a collection of Dataruns, executions of Pipelines generated from the
    Experiment Template over its Signals Set.
    """
    name = fields.StringField(required=True)
    project = fields.StringField()
    template = fields.ReferenceField(Template, reverse_delete_rule=CASCADE)
    dataset = fields.ReferenceField(Dataset, reverse_delete_rule=CASCADE)
    signals = fields.ListField(fields.ReferenceField(Signal, reverse_delete_rule=CASCADE))
    created_by = fields.StringField()

    unique_key_fields = ['name', 'project']

    @property
    def dataruns(self):
        return Datarun.find(experiment=self)


class Datarun(OrionDocument, Status):
    """Datarun object.

    The Datarun object represents a single execution of a Pipeline over a set
    of Signals, within the context of an Experiment.

    It contains all the information about the environment and context where this
    execution took place, which potentially allows to later on reproduce the results
    in a new environment.

    It also contains information about whether the execution was successful or not,
    when it started and ended, and the number of events that were found in this experiment.
    """
    experiment = fields.ReferenceField(Experiment, reverse_delete_rule=CASCADE)
    pipeline = fields.ReferenceField(Pipeline, reverse_delete_rule=CASCADE)
    start_time = fields.DateTimeField()
    end_time = fields.DateTimeField()
    software_versions = fields.ListField(fields.StringField())
    budget_type = fields.StringField()
    budget_amount = fields.IntField()
    num_events = fields.IntField()

    _software_versions = list(freeze.freeze())

    @property
    def signalruns(self):
        return Signalrun.find(datarun=self)

    def start(self):
        """Mark this Datarun as started on DB.

        The ``start_time`` will be set to ``datetime.utcnow()``,
        the ``status`` will be set to RUNNING and the software
        versions will be captured.
        """
        self.start_time = datetime.utcnow()
        self.status = self.STATUS_RUNNING
        self.software_versions = self._software_versions
        self.save()

    def end(self, status):
        """Mark this Datarun as ended on DB.

        The ``end_time`` will be set to ``datetime.utcnow()``, the ``status``
        will be set to the given value, and the ``num_events`` field will be
        populated with the sum of the events detected by the children Signalruns.
        """
        self.end_time = datetime.utcnow()
        self.status = status
        self.num_events = Event.find(signalrun__in=self.signalruns).count()
        self.save()


class Signalrun(OrionDocument, Status):
    """Signalrun object.

    The Signalrun object represents a single executions of a Pipeline on a Signal
    within a Datarun.

    It contains information about whether the execution was successful or not, when
    it started and ended, the number of events that were found by the Pipeline, and
    where the model and metrics are stored.
    """
    datarun = fields.ReferenceField(Datarun, reverse_delete_rule=CASCADE)
    signal = fields.ReferenceField(Signal, reverse_delete_rule=CASCADE)
    start_time = fields.DateTimeField()
    end_time = fields.DateTimeField()
    model_location = fields.StringField()
    metrics_location = fields.StringField()
    num_events = fields.IntField(default=0)

    @property
    def events(self):
        return Event.find(signalrun=self)

    def start(self):
        """Mark this Signalrun as started on DB.

        The ``start_time`` will be set to ``datetime.utcnow()``,
        the ``status`` will be set to RUNNING.
        """
        self.start_time = datetime.utcnow()
        self.status = self.STATUS_RUNNING
        self.save()

    def end(self, status, events):
        """Mark this Signalrun as ended on DB.

        The ``end_time`` will be set to ``datetime.utcnow()``, the ``status``
        will be set to the given value, and the given events will be inserted
        into the Database.
        """
        try:
            if events is None:
                events = []
            for start_time, stop_time, severity in events:
                Event.insert(
                    signalrun=self,
                    signal=self.signal,
                    start_time=start_time,
                    stop_time=stop_time,
                    severity=severity,
                    source=Event.SOURCE_ORION,
                )
        except Exception:
            LOGGER.exception('Error storing signalrun %s events', self.id)
            status = self.STATUS_ERRORED

        self.end_time = datetime.utcnow()
        self.status = status
        self.num_events = len(events)
        self.save()


class Event(OrionDocument):
    """Event object.

    An Event represents an anomalous period on a Signal.
    It is descrived by start and stop times and, optionally, a severity score.

    It is always associated to a Signal and, optionally, to a Signalrun.
    """
    SOURCE_ORION = 'ORION'
    SOURCE_SHAPE_MATCHING = 'SHAPE_MATCHING'
    SOURCE_MANUALLY_CREATED = 'MANUALLY_CREATED'
    SOURCE_CHOICES = (SOURCE_ORION, SOURCE_SHAPE_MATCHING, SOURCE_MANUALLY_CREATED)

    signalrun = fields.ReferenceField(Signalrun, reverse_delete_rule=CASCADE)
    signal = fields.ReferenceField(Signal, reverse_delete_rule=CASCADE)
    start_time = fields.IntField(required=True)
    stop_time = fields.IntField(required=True)
    severity = fields.FloatField()
    source = fields.StringField(required=True, choices=SOURCE_CHOICES)
    num_annotations = fields.IntField(default=0)

    @property
    def annotations(self):
        return Annotation.find(event=self)

    @property
    def event_interactions(self):
        return EventInteraction.find(event=self)

    @property
    def latest_event_interaction(self):
        return self.event_interactions.last()


class EventInteraction(OrionDocument):
    """EventInteraction object.

    The EventInteraction object represents an interaction of a user with an
    Event and includes information about the assocaiated Event, the action
    performed and, if changed, the specified start and stop times.
    """
    event = fields.ReferenceField(Event, reverse_delete_rule=CASCADE)
    action = fields.StringField()
    start_time = fields.IntField(required=True)
    stop_time = fields.IntField(required=True)
    created_by = fields.StringField()


class Annotation(OrionDocument):
    """Annotation object.

    User Annotation related to an event. The Annotations consist on a tag and
    a free text field where the user can insert comments about the Event.
    """
    event = fields.ReferenceField(Event, reverse_delete_rule=CASCADE)
    tag = fields.StringField()
    comment = fields.StringField()
    created_by = fields.StringField()

    def save(self):
        super().save()
        self.event.num_annotations = self.event.annotations.count()
        self.event.save()



"""Orion runner.

Functions responsible for executing Pipelines on Signals
and storing the results inside an Orion Database using
the Orion Explorer.
"""
import logging
import pickle

LOGGER = logging.getLogger(__name__)


def get_outputs_spec(pipeline):
    outputs = ["default"]
    try:
        output_names = pipeline.get_output_names('visualization')
        outputs.append('visualization')
    except ValueError:
        output_names = []

    return outputs, output_names


def process_pipeline_output(orex, signalrun, pipeline_output, output_names):
    if not isinstance(pipeline_output, tuple):
        return pipeline_output

    events = pipeline_output[0]
    if output_names:
        # There might be multiple `default` outputs before the `visualization`
        # outputs in the pipeline_output tuple, thus we get the last entries
        # corresponding to visualization
        visualization = pipeline_output[-len(output_names):]
        visualization_dict = dict(zip(output_names, visualization))
        for name, value in visualization_dict.items():
            kwargs = {
                "filename": '{}-{}.pkl'.format(signalrun.id, name),
                "signalrun_id": signalrun.id,
                "variable": name
            }
            with orex._fs.new_file(**kwargs) as f:
                pickle.dump(value, f)

    return events


def start_signalrun(orex, datarun, signal):
    signalrun = orex.add_signalrun(datarun, signal)
    signalrun.start()
    LOGGER.info('Signalrun %s started', signalrun.id)

    try:
        data = signalrun.signal.load()
        pipeline = signalrun.datarun.pipeline
        mlpipeline = pipeline.load()
        outputs, output_names = get_outputs_spec(mlpipeline)

        LOGGER.info('Running pipeline %s on signal %s', pipeline.name, signal.name)
        pipeline_output = mlpipeline.fit(data, output_=outputs)

        # LOGGER.info('Producing pipeline %s on signal %s', pipeline.name, signal.name)
        # pipeline_output = mlpipeline.predict(data, output_=outputs)

        LOGGER.info('Processing pipeline %s predictions on signal %s', pipeline.name, signal.name)
        events = process_pipeline_output(orex, signalrun, pipeline_output, output_names)
        status = signalrun.STATUS_SUCCESS

    except Exception:
        LOGGER.exception('Signalrun %s crashed', signalrun.id)
        events = []
        status = signalrun.STATUS_ERRORED

    signalrun.end(status, events)


def start_datarun(orex, experiment, pipeline):
    """Start executing a Datarun and store the results on DB.

    Args:
        orex (OrionExplorer):
            OrionExplorer instance to use to store the results
            inside the Database.
        experiment (Experiment or ObjectId or str):
            The Experiment to which the created Datarun will
            belong.
        pipeline (Pipeline or ObjectId or str):
            Pipeline to use for the Datarun.
    """
    datarun = orex.add_datarun(experiment, pipeline)
    datarun.start()
    LOGGER.info('Datarun %s started', datarun.id)

    try:
        for signal in experiment.signals:
            start_signalrun(orex, datarun, signal)

        status = datarun.STATUS_SUCCESS

    except Exception:
        LOGGER.exception('Datarun %s crashed', datarun.id)
        status = datarun.STATUS_ERRORED

    datarun.end(status)



import numpy as np
import pandas as pd
import pytest
from mlblocks import MLPipeline

from orion import analysis


@pytest.fixture
def tadgan_hyperparameters():
    return {
        "mlprimitives.custom.timeseries_preprocessing.time_segments_aggregate#1": {
            "interval": 1,
            "time_column": "timestamp",
        },
        "mlprimitives.custom.timeseries_preprocessing.rolling_window_sequences#1": {
            "target_column": 0,
            "window_size": 100,
            "target_size": 1
        },
        'orion.primitives.tadgan.TadGAN#1': {
            'epochs': 2,
            'verbose': False
        }
    }


@pytest.fixture
def tadgan_pipline(tadgan_hyperparameters):
    pipeline_path = 'tadgan'
    pipline = analysis._load_pipeline(pipeline_path, tadgan_hyperparameters)
    return pipline


def test__build_events_df_empty():
    events = np.array([], dtype=np.float64)

    returned = analysis._build_events_df(events)

    assert returned.empty
    assert list(returned.columns) == ['start', 'end', 'score']


def test__build_events_df_events():
    events = np.array([
        [1.22297040e+09, 1.22299200e+09, 5.72643599e-01],
        [1.22301360e+09, 1.22303520e+09, 5.72643599e-01],
        [1.22351040e+09, 1.22353200e+09, 5.72643599e-01],
    ])

    returned = analysis._build_events_df(events)

    expected = pd.DataFrame({
        'start': [1222970400, 1223013600, 1223510400],
        'end': [1222992000, 1223035200, 1223532000],
        'score': [0.572644, 0.572644, 0.572644]
    })
    pd.testing.assert_frame_equal(returned, expected)


def test__load_pipeline_tadgan(tadgan_pipline, tadgan_hyperparameters):
    hyperset = tadgan_pipline.get_hyperparameters()
    returned = list()
    for k, v in tadgan_hyperparameters.items():
        returned.extend([item in hyperset[k].items() for item in v.items()])

    assert isinstance(tadgan_pipline, MLPipeline)
    assert all(returned)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

try:
    with open('README.md', encoding='utf-8') as readme_file:
        readme = readme_file.read()
except IOError:
    readme = ''

try:
    with open('HISTORY.md', encoding='utf-8') as history_file:
        history = history_file.read()
except IOError:
    history = ''


install_requires = [
    's3fs>=0.2.2,<0.5',
    'baytune>=0.2.3,<0.3',
    'mlblocks>=0.3.0,<0.4',
    'mlprimitives>=0.2.2,<0.3',
    'mongoengine>=0.16.3,<0.17',
    'numpy>=1.15.4,<1.17',
    'pandas>=0.23.4,<0.25',
    'pymongo>=3.7.2,<4',
    'scikit-learn>=0.20.1,<0.21',
    'tabulate>=0.8.3,<0.9',
    'Keras>=2.1.6,<2.4',
    'similaritymeasures>=0.4.3',
    'azure-cognitiveservices-anomalydetector>0.1,<=0.2'
]

setup_requires = [
    'pytest-runner>=2.11.1',
]

tests_require = [
    'pytest>=3.4.2',
    'pytest-cov>=2.6.0',
    'rundoc>=0.4.3,<0.5',
]

development_requires = [
    # general
    'pip>=9.0.1',
    'bumpversion>=0.5.3,<0.6',
    'watchdog>=0.8.3,<0.11',

    # docs
    'm2r>=0.2.0,<0.3',
    'nbsphinx>=0.5.0,<0.7',
    'Sphinx>=1.7.1,<3',
    'sphinx_rtd_theme>=0.2.4,<0.5',
    'autodocsumm>=0.1.10,<1',

    # style check
    'flake8>=3.7.7,<3.8',
    'isort>=4.3.4,<5',

    # fix style issues
    'autoflake>=1.2,<2',
    'autopep8>=1.4.3,<2',

    # distribute on PyPI
    'twine>=1.10.0,<4',
    'wheel>=0.30.0',

    # Advanced testing
    'coverage>=4.5.1,<6',
    'tox>=2.9.1,<4',

    # Documentation style
    'doc8==0.8.0,<0.9',
    'pydocstyle==3.0.0,<4',
]

setup(
    author="MIT Data To AI Lab",
    author_email='dailabmit@gmail.com',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    description="Orion is a machine learning library built for data generated by satellites.",
    entry_points={
        'console_scripts': [
            'orion=orion.cli:main'
        ],
        'mlblocks': [
            'primitives=orion:MLBLOCKS_PRIMITIVES',
            'pipelines=orion:MLBLOCKS_PIPELINES'
        ],
    },
    extras_require={
        'test': tests_require,
        'dev': development_requires + tests_require,
    },
    include_package_data=True,
    install_requires=install_requires,
    keywords='orion',
    license="MIT license",
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/markdown',
    name='orion-ml',
    packages=find_packages(include=['orion', 'orion.*']),
    python_requires='>=3.6,<3.8',
    setup_requires=setup_requires,
    test_suite='tests',
    tests_require=tests_require,
    url='https://github.com/D3-AI/Orion',
    version='0.1.3.dev0',
    zip_safe=False,
)



"""Orion Database subpackage.

This subpackage contains all the code related to the
Orion Database usage.
"""
from orion.db import schema
from orion.db.explorer import OrionDBExplorer

__all__ = (
    'OrionDBExplorer',
    'schema',
)



"""Orion Database Schema.

This module contains some utility functions and classes that assist
on the usage of MongoEngine to define a DB Schema.
"""

import copy
from datetime import datetime

import pandas as pd
from bson import ObjectId
from mongoengine import Document, fields
from mongoengine.base.metaclasses import TopLevelDocumentMetaclass


def walk(document, transform):
    if not isinstance(document, dict):
        return document

    new_doc = dict()
    for key, value in document.items():
        if isinstance(value, dict):
            value = walk(value, transform)
        elif isinstance(value, list):
            value = [walk(v, transform) for v in value]

        new_key, new_value = transform(key, value)
        new_doc[new_key] = new_value

    return new_doc


def remove_dots(document):
    return walk(document, lambda key, value: (key.replace('.', '-'), value))


def restore_dots(document):
    return walk(document, lambda key, value: (key.replace('-', '.'), value))


def _merge_meta(base, child):
    """Merge the base and the child meta attributes.

    List entries, such as ``indexes`` are concatenated.
    ``abstract`` value is set to ``True`` only if defined as such
    in the child class.

    Args:
        base (dict):
            ``meta`` attribute from the base class.
        child (dict):
            ``meta`` attribute from the child class.

    Returns:
        dict:
            Merged metadata.
    """
    base = copy.deepcopy(base)
    child.setdefault('abstract', False)
    for key, value in child.items():
        if isinstance(value, list):
            base.setdefault(key, []).extend(value)
        else:
            base[key] = value

    return base


class OrionMeta(TopLevelDocumentMetaclass):
    """Metaclass for the OrionDocument class.

    It ensures that the ``meta`` attribute from the OrionDocument
    parent class is used even if the child class defines a new one
    by merging both of them together.
    """

    def __new__(mcs, name, bases, attrs):
        if 'meta' in attrs:
            meta = attrs['meta']
            for base in bases:
                if base is not Document and hasattr(base, '_meta'):
                    meta = _merge_meta(base._meta, meta)

            attrs['meta'] = meta

        if 'unique_key_fields' in attrs:
            indexes = attrs.setdefault('meta', {}).setdefault('indexes', [])
            indexes.append({
                'fields': attrs['unique_key_fields'],
                'unique': True,
                'sparse': True,
            })

        return super().__new__(mcs, name, bases, attrs)


class OrionDocument(Document, metaclass=OrionMeta):
    """Parent class for all the Document classes in Orion.

    This class defines a few defaults, such as the ``instert_time`` field
    and index, as well as a few utility methods to ease the interaction
    with the database models.
    """

    insert_time = fields.DateTimeField(default=datetime.utcnow)

    meta = {
        'indexes': [
            '$insert_time',
        ],
        'abstract': True,
        'index_background': True,
        'auto_create_index': True,
    }

    @staticmethod
    def _get_id(obj):
        if isinstance(obj, ObjectId):
            return obj
        elif isinstance(obj, Document):
            return obj.id
        elif obj:
            return ObjectId(obj)

    @classmethod
    def find(cls, as_df_=False, only_=None, exclude_=None, **kwargs):
        name = cls.__name__.lower()
        if name in kwargs:
            kwargs['id'] = cls._get_id(kwargs.pop(name))

        query = {
            key: value
            for key, value in kwargs.items()
            if value is not None
        }
        cursor = cls.objects(**query)
        if only_:
            cursor = cursor.only(*only_)

        if exclude_:
            cursor = cursor.exclude(*exclude_)

        if not as_df_:
            return cursor

        df = pd.DataFrame([
            document.to_mongo()
            for document in cursor
        ]).rename(columns={'_id': name + '_id'})

        if exclude_:
            for column in exclude_:
                if column in df:
                    del df[column]

        return df

    @classmethod
    def get(cls, **kwargs):
        query = {
            key: value
            for key, value in kwargs.items()
            if value is not None
        }
        if not query:
            raise ValueError('Empty queries not supported')

        cursor = cls.find(as_df_=False, **query)
        if not cursor:
            raise ValueError('No {} found for query {}'.format(cls.__name__, query))
        elif cursor.count() > 1:
            raise ValueError('Multiple {}s found for query {}'.format(cls.__name__, query))

        return cursor.first()

    @classmethod
    def find_one(cls, **kwargs):
        return cls.find(**kwargs).first()

    @classmethod
    def last(cls, **kwargs):
        return cls.find(**kwargs).order_by('-insert_time').first()

    @classmethod
    def insert(cls, **kwargs):
        document = cls(**kwargs)
        document.save()

        return document

    @classmethod
    def find_or_insert(cls, **kwargs):
        document = cls.find_one(**kwargs)
        if document is None:
            document = cls.insert(**kwargs)

        return document


class Status:
    """Mixin that adds a status field and a method to check its live value."""

    status = fields.StringField()

    STATUS_PENDING = 'PENDING'
    STATUS_RUNNING = 'RUNNING'
    STATUS_SUCCESS = 'SUCCESS'
    STATUS_ERRORED = 'ERRORED'

    def get_status(self):
        self.reload()
        return self.status


def key_has_dollar(d):
    """Recursively check if any key in a dict contains a dollar sign."""
    for k, v in d.items():
        if k.startswith('$') or (isinstance(v, dict) and key_has_dollar(v)):
            return True


class PipelineField(fields.DictField):

    def to_mongo(self, value, use_db_field=True, fields=None):
        value = remove_dots(value)
        return super().to_mongo(value, use_db_field, fields)

    def to_python(self, value):
        value = restore_dots(value)
        return super().to_python(value)

    def validate(self, value):
        """Make sure that a list of valid fields is being used."""
        if not isinstance(value, dict):
            self.error('Only dictionaries may be used in a PipelineField')

        if fields.key_not_string(value):
            msg = ('Invalid dictionary key - documents must '
                   'have only string keys')
            self.error(msg)

        if key_has_dollar(value):
            self.error('Invalid dictionary key name - keys may not start with '
                       '"$" character')

        super(fields.DictField, self).validate(value)



import numpy as np


class MeanEstimator:
    """Mean Estimator.

    This is a dummy estimator that always returns a constant value,
    which consist on the mean value from the given input.

    This estimator is here only to serve as reference of what
    an estimator primitive looks like, and is not intended to be
    used in real scenarios.
    """

    def __init__(self, value_column='value'):
        self._value_column = value_column

    def fit(self, X):
        values = X[self._value_column]
        self._mean = np.mean(values)

    def predict(self, X):
        return np.full(len(X), self._mean)
